"""This subpackage contains the classes for defining the target in SRIM simulations."""

from irradiapy.srimpy.target.element import Element
from irradiapy.srimpy.target.layer import Layer
from irradiapy.srimpy.target.target import Target



"""This module contains the `Layer` class."""

from dataclasses import dataclass, field

from irradiapy.srimpy.target.element import Element


@dataclass
class Layer:
    """Class for defining a layer in SRIM simulations.

    Attributes
    ----------
    width : float
        The thickness of the layer in angstroms.
    phase : int
        The phase of the layer. 0 = solid, 1 = gas.
    density : float
        The density of the layer in g/cm^3.
    bragg : int = 1 (default)
        Stopping corrections for special bonding in compound targets.
    elements : list[Element]
        The list of elements in the layer.
    stoichs : list[float]
        The list of stoichiometries for the elements in the layer.
    """

    width: float
    phase: int
    density: float
    elements: list[Element]
    stoichs: list[float]
    bragg: int = 1
    nelements: int = field(init=False)

    def __post_init__(self) -> None:
        self.nelements = len(self.elements)
        self.nelements = len(self.elements)



"""Utility functions related to injected ions for SRIM data."""

from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from irradiapy.math_utils import fit_gaussian
from irradiapy.srimpy.srimdb import SRIMDB


def plot_injected(
    srimdb: SRIMDB,
    bins: int = 100,
    plot_path: Optional[Path] = None,
    dpi: int = 300,
    path_fit: Optional[Path] = None,
    p0: Optional[float] = None,
    asymmetry: float = 1.0,
) -> None:
    """Plot injected ions final depth distribution.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    bins : int, optional
        Depth bins, by default 100.
    plot_path : Path, optional
        Output path for the plot, by default None. If None, the plot is shown.
    dpi : int, optional
        Dots per inch, by default 300.
    path_fit : Path, optional
        Output path for the fit parameters, by default None.
    p0 : float, optional
        Initial guess of fit parameters, by default None.
    asymmetry : float, optional
        Asymmetry fit parameter bound, by default 1.0.
    """
    # Read
    depths = np.array([ion[0] for ion in srimdb.range3d.read(what="depth")])
    # Histogram
    histogram, depth_edges = np.histogram(depths, bins=bins)
    depth_centers = (depth_edges[:-1] + depth_edges[1:]) / 2.0
    # Fit
    fit, injected_fit = False, None
    if path_fit:
        try:
            popt, _, injected_fit = fit_gaussian(
                depth_centers, histogram, p0, asymmetry
            )
            if path_fit:
                with open(path_fit, "w", encoding="utf-8") as file:
                    file.write("Injected atoms gaussian fit: z0, sigma, A, a\n")
                    file.write(
                        (
                            "See Eq. (1) of Nuclear Instruments and Methods in Physics "
                            "Research B 500-501 (2021) 52-56\n"
                        )
                    )
                    file.write(", ".join(map(str, popt)) + "\n")
            fit = True
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Fit failed: {exc}")
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec()
    # Scatter
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_ylabel("Counts per ion")
    ax.scatter(depth_centers, histogram)
    if fit:
        ax.plot(
            depth_centers,
            injected_fit(depth_centers),
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
    # Finish
    fig.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()



"""Cluster analysis module."""

import sqlite3
from pathlib import Path
from typing import Optional

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec
from matplotlib.image import NonUniformImage
from matplotlib.ticker import MaxNLocator
from scipy.optimize import curve_fit

from irradiapy import dtypes
from irradiapy.io.xyzreader import XYZReader
from irradiapy.io.xyzwriter import XYZWriter


def clusterize(defects: np.ndarray, cutoff: float) -> tuple[np.ndarray, np.ndarray]:
    """Finds defects clusters in the given defects.

    Notes
    -----
    Algorithm: union-find path compression. Output data structure (atom clusters):
    ``{1: [[accx0, ...], [accy0, ...], [accz0, ...]],
    2: [[accx0, ...], [accy0, ...], [accz0, ...]]}``

    Parameters
    ----------
    defects : dtypes.DEFECT_CHECK
        The defects.
    cutoff : float
        Cutoff distance for clustering.

    Returns
    -------
    tuple[dtypes.ACLUSTER_CHECK, dtypes.OCLUSTER_CHECK]
        Atomic clusters and object clusters.
    """
    ndefects = len(defects)
    defects_ = np.zeros(ndefects, dtype=dtypes.defect)
    defects_["type"] = defects["type"]
    if (
        "x" in defects.dtype.names
        or "y" in defects.dtype.names
        or "z" in defects.dtype.names
    ):
        defects_["pos"][:, 0] = defects["x"]
        defects_["pos"][:, 1] = defects["y"]
        defects_["pos"][:, 2] = defects["z"]
        defects = defects_

    cutoff2 = cutoff**2
    natoms = defects.size
    aclusters = np.empty(natoms, dtype=dtypes.acluster)
    aclusters["pos"] = defects["pos"]
    aclusters["type"] = defects["type"]
    aclusters["cluster"] = np.arange(1, natoms + 1)

    for i in range(natoms):
        curr_cluster = aclusters[i]["cluster"]
        dists = np.sum(
            np.square(aclusters[i]["pos"] - aclusters[i + 1 :]["pos"]), axis=1
        )
        neighbourhood = aclusters["cluster"][np.where(dists <= cutoff2)[0] + i + 1]
        if neighbourhood.size:
            for neighbour in neighbourhood:
                if neighbour != curr_cluster:
                    aclusters["cluster"][
                        aclusters["cluster"] == neighbour
                    ] = curr_cluster

    nclusters = np.unique(aclusters["cluster"])
    for i in range(nclusters.size):
        aclusters["cluster"][aclusters["cluster"] == nclusters[i]] = i + 1

    oclusters = atom_to_object(aclusters)
    return aclusters, oclusters


def atom_to_object(aclusters: np.ndarray) -> np.ndarray:
    """Transform atom clusters into object clusters.

    Parameters
    ----------
    aclusters : dtypes.ACLUSTER_CHECK
        Atomic clusters.

    Returns
    -------
    dtypes.OCLUSTER_CHECK
        Object clusters.
    """
    nclusters = np.unique(aclusters["cluster"])
    oclusters = np.empty(nclusters.size, dtype=dtypes.ocluster)
    for i in range(nclusters.size):
        acluster = aclusters[aclusters["cluster"] == nclusters[i]]
        oclusters[i]["pos"] = np.mean(acluster["pos"], axis=0)
        oclusters[i]["type"] = acluster[0]["type"]
        oclusters[i]["size"] = acluster.size
    return oclusters


def clusterize_file(
    path_collisions: Path,
    cutoff_sia: float,
    cutoff_vac: float,
    path_aclusters: Path,
    path_oclusters: Path,
    irradiated_particle: str = "Projectile",
) -> None:
    """Finds defect clusters in the given file.

    Parameters
    ----------
    path_collisions : Path
        Path of the file where defects are.
    cutoff_sia : float
        Cutoff distance for interstitials clustering.
    cutoff_vac : float
        Cutoff distance for vacancies clustering.
    path_aclusters : Path
        Where atomic clusters will be stored.
    path_oclusters : Path
        Where object clusters will be stored.
    irradiated_particle : str, optional
        Name of the irradiated particle, by default "Projectile".
    """
    reader = XYZReader(path_collisions)
    nsim = 0
    with XYZWriter(path_aclusters) as awriter, XYZWriter(path_oclusters) as owriter:
        for defects in reader:
            nsim += 1
            cond = defects["type"] == 0
            sia, vac = defects[~cond], defects[cond]
            iaclusters, ioclusters = clusterize(sia, cutoff_sia)
            vaclusters, voclusters = clusterize(vac, cutoff_vac)
            aclusters = np.concatenate((iaclusters, vaclusters))
            oclusters = np.concatenate((ioclusters, voclusters))
            awriter.save(aclusters, extra_comment=f"{irradiated_particle}={nsim}")
            owriter.save(oclusters, extra_comment=f"{irradiated_particle}={nsim}")


# region Analysis
def __get_size_depth_histogram(
    sizes: np.ndarray,
    depths: np.ndarray,
    depth_bins: int = 100,
    size_bins: Optional[np.integer] = None,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
]:
    """Generate a 2D histogram of sizes and depths.

    Parameters
    ----------
    sizes : dtypes.FLOAT_CHECK
        An array of size values.
    depths : dtypes.FLOAT_CHECK
        An array of depth values.
    depth_bins : int, optional
        Number of bins for the depth axis, by default 100.
    size_bins : int, optional
        Number of bins for the size axis, by default None (one bin per size).

    Returns
    -------
    tuple[dtypes.FLOAT_CHECK, dtypes.FLOAT_CHECK, dtypes.FLOAT_CHECK, dtypes.FLOAT_CHECK,
    dtypes.FLOAT_CHECK]
        A tuple containing: the values of the histogram, the bin edges of the size histogram,
        the bin edges of the depth histogram, the centers of the size bins, the centers of the
        depth bins.
    """
    max_size = sizes.max()
    if size_bins is None:
        size_bins = max_size
    histogram, size_edges, depth_edges = np.histogram2d(
        x=sizes,
        y=depths,
        bins=(size_bins, depth_bins),
        range=((0.5, max_size + 0.5), (depths.min(), depths.max())),
    )
    size_centers = (size_edges[:-1] + size_edges[1:]) / 2.0
    depth_centers = (depth_edges[:-1] + depth_edges[1:]) / 2.0
    return histogram, size_edges, depth_edges, size_centers, depth_centers


def __get_size_histogram(
    sizes: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generate a histogram of cluster sizes.

    Parameters
    ----------
    sizes : dtypes.FLOAT_CHECK
        An array of cluster sizes.

    Returns
    -------
    tuple[dtypes]
        A tuple containing: the histogram of cluster sizes, the edges of the histogram bins and
        the centers of the histogram bins.
    """
    max_size = sizes.max()
    histogram, size_edges = np.histogram(
        sizes, bins=max_size, range=(0.5, max_size + 0.5)
    )
    size_centers = (size_edges[:-1] + size_edges[1:]) / 2.0
    return histogram, size_edges, size_centers


def get_clusters(path_oclusters: Path, db_path: Path) -> None:
    """Processes cluster data from an XYZ file and stores the results in a SQLite database.

    Parameters
    ----------
    path_oclusters : Path
        Path to the input XYZ file containing cluster data.
    db_path : Path
        Path to the output SQLite database file.
    """
    isizes, vsizes = np.array([], dtype=int), np.array([], dtype=int)
    idepths, vdepths = np.array([], dtype=float), np.array([], dtype=float)

    reader = XYZReader(path_oclusters)
    for oclusters in reader:
        cond = oclusters["type"] == 0
        vacancies = oclusters[cond]
        if vacancies.size > 0:
            vsizes = np.concatenate((vsizes, vacancies["size"]))
            vdepths = np.concatenate((vdepths, vacancies["pos"][:, 0]))
        interstitials = oclusters[~cond]
        if interstitials.size > 0:
            isizes = np.concatenate((isizes, interstitials["size"]))
            idepths = np.concatenate((idepths, interstitials["pos"][:, 0]))

    con = sqlite3.connect(db_path)
    cur = con.cursor()
    cur.execute("DROP TABLE IF EXISTS interstitials")
    cur.execute("CREATE TABLE interstitials (size INTEGER, depth REAL)")
    cur.executemany(
        "INSERT INTO interstitials (size, depth) VALUES (?, ?)",
        np.column_stack((isizes, idepths)),
    )
    cur.execute("DROP TABLE IF EXISTS vacancies")
    cur.execute("CREATE TABLE vacancies (size INTEGER, depth REAL)")
    cur.executemany(
        "INSERT INTO vacancies (size, depth) VALUES (?, ?)",
        np.column_stack((vsizes, vdepths)),
    )
    con.commit()
    cur.close()
    con.close()


def load_results(
    db_path: Path, what: str = "*", condition: str = ""
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Load results from an SQLite database generated with `get_clusters`.

    Parameters
    ----------
    db_path : Path
        The path to the SQLite database file.
    what : str, optional
        The columns to select from the tables, by default "*".
    condition : str, optional
        An optional SQL condition to filter the results, by default "".

    Returns
    -------
    tuple[dtypes.INT_CHECK, dtypes.FLOAT_CHECK, dtypes.INT_CHECK, dtypes.FLOAT_CHECK]
        A tuple containing: the sizes of the interstitials, the depths of the interstitials, the
        sizes of the vacancies, the depths of the vacancies
    """
    con = sqlite3.connect(db_path)
    cur = con.cursor()
    cur.execute(f"SELECT {what} FROM interstitials {condition}")
    data = np.array(cur.fetchall())
    isizes = data[:, 0].astype(int)
    idepths = data[:, 1]
    cur.execute(f"SELECT {what} FROM vacancies {condition}")
    data = np.array(cur.fetchall())
    vsizes = data[:, 0].astype(int)
    vdepths = data[:, 1]
    return isizes, idepths, vsizes, vdepths


def __scaling_law_fit(
    centers: np.ndarray, counts: np.ndarray
) -> tuple[float, float, callable]:
    """Fit a scaling law to the given histogram data.

    Parameters
    ----------
    centers : dtypes.FLOAT_CHECK
        The centers of the bins.
    counts : dtypes.FLOAT_CHECK
        The values of the histogram.

    Returns
    -------
    tuple
        A tuple containing: the prefactor of the scaling law, the exponent of the scaling law,
        and the scaling law function.
    """
    # pylint: disable=unbalanced-tuple-unpacking
    popt, _ = curve_fit(lambda x, a, b: a + b * x, np.log10(centers), np.log10(counts))
    a, s = popt
    a, s = 10.0**a, -s

    def curve(x):
        return a / x**s

    return a, s, curve


def plot_results(
    db_path: Path,
    out_dir: Path,
    nions: int,
    depth_bins: int = 100,
    size_bins: Optional[int] = None,
    min_depth: Optional[float] = None,
    max_depth: Optional[float] = None,
) -> None:
    """Plots the results of size-depth histograms, clustering fractions,
    and scaling laws for interstitials and vacancies.

    Parameters
    ----------
    db_path : Path
        The path to the SQLite database file.
    out_dir : Path
        The output directory where the plots and results will be saved.
    nions : int
        The number of ions used for normalization.
    depth_bins : int, optional
        The number of bins for depth histograms, by default 100.
    size_bins : int, optional
        The number of bins for size histograms, by default None (one bin per size).
    """
    isizes, idepths, vsizes, vdepths = load_results(db_path)

    # Create masks for depth filtering
    imask = np.ones_like(idepths, dtype=bool)
    vmask = np.ones_like(vdepths, dtype=bool)
    if min_depth is not None:
        imask &= idepths >= min_depth
        vmask &= vdepths >= min_depth
    if max_depth is not None:
        imask &= idepths <= max_depth
        vmask &= vdepths <= max_depth
    isizes, idepths = isizes[imask], idepths[imask]
    vsizes, vdepths = vsizes[vmask], vdepths[vmask]

    (
        sia_histogram,
        sia_size_edges,
        sia_depth_edges,
        sia_size_centers,
        sia_depth_centers,
    ) = __get_size_depth_histogram(
        isizes, idepths, depth_bins=depth_bins, size_bins=size_bins
    )
    (
        vac_histogram,
        vac_size_edges,
        vac_depth_edges,
        vac_size_centers,
        vac_depth_centers,
    ) = __get_size_depth_histogram(
        vsizes, vdepths, depth_bins=depth_bins, size_bins=size_bins
    )
    sia_histogram /= nions
    vac_histogram /= nions

    # Plot size-depth
    fig = plt.figure()
    # Interstitials
    iax = plt.subplot2grid((9, 2), (1, 0), rowspan=8)
    iax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
    iax.set_xlabel("Self-interstitial size")
    iax.set_xlim(sia_size_edges[[0, -1]])
    iax.set_ylim(sia_depth_edges[[0, -1]])
    iim = NonUniformImage(
        iax,
        interpolation="nearest",
        norm="log",
        extent=(*sia_size_edges[[0, -1]], *sia_depth_edges[[0, -1]]),
    )
    iim.set_data(sia_size_centers, sia_depth_centers, sia_histogram.T)
    iax.add_image(iim)
    # Vacancies
    vax = plt.subplot2grid((9, 2), (1, 1), rowspan=8)
    vax.yaxis.set_label_position("right")
    vax.yaxis.tick_right()
    # vax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
    vax.set_yticklabels([])
    vax.set_xlabel("Vacancy size")
    vax.set_xlim(vac_size_edges[[0, -1]])
    vax.set_ylim(vac_depth_edges[[0, -1]])
    vim = NonUniformImage(
        vax,
        interpolation="nearest",
        norm="log",
        extent=(*vac_size_edges[[0, -1]], *vac_depth_edges[[0, -1]]),
    )
    vim.set_data(vac_size_centers, vac_depth_centers, vac_histogram.T)
    vax.add_image(vim)
    # Color bar
    cax = plt.subplot2grid((9, 2), (0, 0), colspan=2)
    cbar = fig.colorbar(vim, cax=cax, orientation="horizontal")
    cbar.set_label("Counts per ion")
    cbar.ax.xaxis.set_ticks_position("top")
    cbar.ax.xaxis.set_label_position("top")
    # Final touches
    plt.tight_layout()
    if out_dir:
        plt.savefig(out_dir / "clusters_depth.png", dpi=300)
    else:
        plt.show()
    plt.close()

    # Clustering fraction
    # One bin, one size
    if size_bins is not None:
        (
            sia_histogram,
            sia_size_edges,
            sia_depth_edges,
            sia_size_centers,
            sia_depth_centers,
        ) = __get_size_depth_histogram(
            isizes, idepths, depth_bins=depth_bins, size_bins=None
        )
        (
            vac_histogram,
            vac_size_edges,
            vac_depth_edges,
            vac_size_centers,
            vac_depth_centers,
        ) = __get_size_depth_histogram(
            vsizes, vdepths, depth_bins=depth_bins, size_bins=None
        )
    sia_histogram /= nions
    vac_histogram /= nions
    # Interstitials
    ihistogram1 = sia_histogram.copy()
    ihistogram1 *= sia_size_centers[:, np.newaxis]
    imonomoer = ihistogram1[0]
    itotal = ihistogram1.sum(axis=0)
    # Vacancies
    vhistogram1 = vac_histogram.copy()
    vhistogram1 *= vac_size_centers[:, np.newaxis]
    vmonomoer = vhistogram1[0]
    vtotal = vhistogram1.sum(axis=0)
    # Plot
    plt.scatter(sia_depth_centers, 1.0 - imonomoer / itotal, label="SIA")
    plt.scatter(vac_depth_centers, 1.0 - vmonomoer / vtotal, label="VAC")
    plt.xlabel(r"Depth ($\mathrm{\AA}$)")
    plt.ylabel("Clustering fraction")
    plt.legend()
    plt.tight_layout()
    if out_dir:
        plt.savefig(out_dir / "clustering_fraction.png", dpi=300)
    else:
        plt.show()
    plt.close()

    # Plot interstitials scaling law
    try:
        icounts = sia_histogram.sum(axis=1)
        icounts_ = icounts[icounts > 0]
        sia_size_centers_ = sia_size_centers[icounts > 0]
        small_counts, big_counts = icounts_[:10], icounts_[10:]
        small_centers, big_centers = sia_size_centers_[:10], sia_size_centers_[10:]
        sia_small_a, sia_small_s, sia_small_curve = __scaling_law_fit(
            small_centers, small_counts
        )
        sia_big_a, sia_big_s, sia_big_curve = __scaling_law_fit(big_centers, big_counts)
        # Plot
        plt.scatter(
            small_centers,
            small_counts,
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
        plt.plot(
            small_centers,
            sia_small_curve(small_centers),
            label=r"Small self-interstitials",
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
        plt.scatter(
            big_centers,
            big_counts,
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][1],
        )
        plt.plot(
            big_centers,
            sia_big_curve(big_centers),
            label=r"Big self-interstitials",
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][1],
        )
        plt.yscale("log")
        plt.xscale("log")
        plt.xlabel("Size")
        plt.ylabel("Counts per ion")
        plt.legend()
        plt.tight_layout()
        if out_dir:
            plt.savefig(out_dir / "clusters_scaling_law_interstitials.png", dpi=300)
        else:
            plt.show()
        plt.close()
    except Exception:  # pylint: disable=broad-exception-caught
        print(
            "Error fitting interstitials scaling law. This functionality is a "
            "general recipe and may not work for all cases."
        )

    # Plot vacancies scaling law
    try:
        vcounts = vac_histogram.sum(axis=1)
        vcounts_ = vcounts[vcounts > 0]
        vac_size_centers_ = vac_size_centers[vcounts > 0]
        small_counts, big_counts = vcounts_[:10], vcounts_[10:]
        small_centers, big_centers = vac_size_centers_[:10], vac_size_centers_[10:]
        vac_small_a, vac_small_s, vac_small_curve = __scaling_law_fit(
            small_centers, small_counts
        )
        vac_big_a, vac_big_s, vac_big_curve = __scaling_law_fit(big_centers, big_counts)
        # Plot
        plt.scatter(
            small_centers,
            small_counts,
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
        plt.plot(
            small_centers,
            vac_small_curve(small_centers),
            label=r"Small vacancies",
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
        plt.scatter(
            big_centers,
            big_counts,
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][1],
        )
        plt.plot(
            big_centers,
            vac_big_curve(big_centers),
            label=r"Big vacancies",
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][1],
        )
        plt.yscale("log")
        plt.xscale("log")
        plt.xlabel("Size")
        plt.ylabel("Counts per ion")
        plt.legend()
        plt.tight_layout()
        if out_dir:
            plt.savefig(out_dir / "clusters_scaling_law_vacancies.png", dpi=300)
            with open(
                out_dir / "clusters_scaling_law.txt", "w", encoding="utf-8"
            ) as file:
                file.write("Scaling law: A/x**S; A, S\n")
                file.write(f"Small SIA {sia_small_a}, {sia_small_s}\n")
                file.write(f"Big SIA {sia_big_a}, {sia_big_s}\n")
                file.write(f"Small VAC {vac_small_a}, {vac_small_s}\n")
                file.write(f"Big VAC {vac_big_a}, {vac_big_s}\n")
        else:
            plt.show()
        plt.close()
    except Exception:  # pylint: disable=broad-exception-caught
        print(
            "Error fitting vacancies scaling law. This functionality is a "
            "general recipe and may not work for all cases."
        )


def mddb_analysis(
    debris_dir: Path,
    cutoff_sia: float,
    cutoff_vac: float,
    bin_width: int = 10,
    lognorm: bool = True,
    plot_path: Optional[Path] = None,
    dpi: int = 300,
) -> None:
    """Analyzes molecular dynamics debris data to generate histograms of cluster sizes
    for interstitials and vacancies.

    Parameters
    ----------
    debris_dir : Path
        Directory containing debris data files.
    cutoff_sia : float
        Cutoff distance for clustering self-interstitial atoms (SIAs).
    cutoff_vac : float
        Cutoff distance for clustering vacancies.
    bin_width : int, optional
        Bin size for histogram. Defaults to 10.
    lognorm : bool, optional
        Use log normalization for the color scale. Defaults to True.
    plot_path : Path, optional
        Path to save the plot. Defaults to None. If None, the plot is shown.
    dpi : int, optional
        DPI of the plot. Defaults to 300.
    """
    epkas, clusters, nfiles = [], {}, {}
    for file_path in debris_dir.iterdir():
        if file_path.is_dir():
            epka = int(file_path.name)
            epkas.append(epka)
            clusters[epka] = {"sia": [], "vac": []}
            nfiles[epka] = 0
            for file_path in file_path.iterdir():
                defects = list(XYZReader(file_path))[0]
                cond = defects["type"] == 0
                vacancies = defects[cond]
                interstitials = defects[np.invert(cond)]
                _, ioclusters = clusterize(interstitials, cutoff_sia)
                _, voclusters = clusterize(vacancies, cutoff_vac)
                clusters[epka]["sia"] += ioclusters["size"].tolist()
                clusters[epka]["vac"] += voclusters["size"].tolist()
                nfiles[epka] += 1
    epkas.sort()
    # Energy binning
    nepka_bins = len(epkas)
    # Determine maximum cluster size
    max_sia, max_vac = 0, 0
    for siasvacs in clusters.values():
        max_sia0 = max(siasvacs["sia"])
        max_vac0 = max(siasvacs["vac"])
        max_sia = max_sia0 if max_sia0 > max_sia else max_sia
        max_vac = max_vac0 if max_vac0 > max_vac else max_vac
    # Round to closest highest multiple of bin_width starting from 10,
    # where binning changes
    # if bin_width = 10, then 83 > 90
    # if bin_width = 10, then 57 > 60
    max_sia = 10 + int(np.ceil((max_sia - 10) / bin_width) * bin_width)
    max_vac = 10 + int(np.ceil((max_vac - 10) / bin_width) * bin_width)
    # Use the same maximum size for both types of clusters
    if max_vac > max_sia:
        max_sia = max_vac
    else:
        max_vac = max_sia
    # Histogram bin edges
    sia_edges = np.concatenate(
        (np.arange(0.5, 10.5), np.arange(10.5, max_sia + bin_width + 0.5, bin_width))
    )
    vac_edges = np.concatenate(
        (np.arange(0.5, 10.5), np.arange(10.5, max_vac + bin_width + 0.5, bin_width))
    )
    # Histogram
    ihist = []
    vhist = []
    for epka in epkas:
        siasvacs = clusters[epka]
        hist, _ = np.histogram(
            siasvacs["sia"],
            bins=sia_edges,
            # weights=siasvacs["sia"],
            density=False,
        )
        ihist.append(hist / nfiles[epka])
        hist, _ = np.histogram(
            siasvacs["vac"],
            bins=vac_edges,
            # weights=siasvacs["vac"],
            density=False,
        )
        vhist.append(hist / nfiles[epka])
    ihist = np.array(ihist)  # row: energy, column: size
    vhist = np.array(vhist)  # row: energy, column: size
    # Mask to remove zero counts.
    ihist = np.ma.masked_where(ihist == 0, ihist)
    vhist = np.ma.masked_where(vhist == 0, vhist)
    vmin = min(ihist.min(), vhist.min())
    vmax = max(ihist.max(), vhist.max())
    # Norm definition
    if lognorm:
        norm = mcolors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
    # Plot
    fig = plt.figure(figsize=(12, 6))
    gs = GridSpec(1, 3, width_ratios=[1, 1, 0.05])

    iax = fig.add_subplot(gs[0, 0])
    iim = iax.imshow(
        ihist.T,
        origin="lower",
        aspect="auto",
        norm=norm,
        cmap="viridis",
        extent=[0, nepka_bins, 0, len(sia_edges) - 1],
    )
    iax.set_title("Interstitials")
    iax.set_xlabel("Energy (keV)")
    iax.set_ylabel("Cluster size")

    vax = fig.add_subplot(gs[0, 1])
    vax.imshow(
        vhist.T,
        origin="lower",
        aspect="auto",
        norm=norm,
        cmap="viridis",
        extent=[0, nepka_bins, 0, len(vac_edges) - 1],
    )
    vax.set_title("Vacancies")
    vax.set_xlabel("Energy (keV)")
    vax.set_yticklabels([])
    vax.yaxis.set_major_locator(MaxNLocator(integer=True))

    # Set correct energy labels and ticks
    for ax in [iax, vax]:
        ax.set_xticks(np.arange(nepka_bins) + 0.5)
        ax.set_xticklabels([f"{int(e/1e3)}" for e in epkas])
    # SIA size labels: 1, 2, ..., 10, 11-20, ...
    if bin_width == 1:
        labels = [str(i) for i in range(1, max_sia + 1)]
    else:
        labels = [str(i) for i in range(1, 11)] + [
            f"{i}-{i+bin_width-1}" for i in range(11, max_sia, bin_width)
        ]
    iax.yaxis.set_major_locator(MaxNLocator(integer=True))
    iax.set_yticks(np.arange(len(sia_edges) - 1) + 0.5)
    iax.set_yticklabels(labels)
    # VAC size labels: 1, 2, ..., 10, 11-20, ...
    if bin_width == 1:
        labels = [str(i) for i in range(1, max_vac + 1)]
    else:
        labels = [str(i) for i in range(1, 11)] + [
            f"{i}-{i+bin_width-1}" for i in range(11, max_vac, bin_width)
        ]
    vax.yaxis.set_major_locator(MaxNLocator(integer=True))
    vax.set_yticks(np.arange(len(vac_edges) - 1) + 0.5)
    # vax.set_yticklabels(labels)
    # vax.yaxis.tick_right()

    cax = fig.add_subplot(gs[0, 2])
    fig.colorbar(iim, cax=cax, label="Mean number of clusters")

    plt.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()


# endregion


# region Counting
def write_count_0d(
    path_oclusters: Path,
    out_dir: Optional[Path] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Processes clusters from a given file, separates them into vacancies and interstitials,
    computes histograms of their sizes, and optionally writes the histograms to a CSV file.

    Parameters
    ----------
    path_oclusters : Path
        Path to the input XYZ file containing cluster data.
    out_dir : Path, optional
        Directory where the histograms will be saved, by default None.

    Returns
    -------
    tuple[dtypes.INT_CHECK, dtypes.INT_CHECK]
        A tuple containing: the sizes of the interstitials and the sizes of the vacancies.
    """
    isizes, vsizes = np.array([], dtype=int), np.array([], dtype=int)
    for oclusters in XYZReader(path_oclusters):
        cond = oclusters["type"] == 0
        vacancies = oclusters[cond]
        if vacancies.size > 0:
            vsizes0 = vacancies["size"]
            vsizes = np.concatenate((vsizes, vsizes0))
        cond = np.invert(cond)
        interstitials = oclusters[cond]
        if interstitials.size > 0:
            isizes0 = interstitials["size"]
            isizes = np.concatenate((isizes, isizes0))
    sia_histogram, _, _ = __get_size_histogram(isizes)
    vac_histogram, _, _ = __get_size_histogram(vsizes)
    if out_dir:
        out_path = out_dir / "oclusters0D.csv"
        with open(out_path, "w", encoding="utf-8") as file:
            file.write(",".join(map(str, sia_histogram)))
            file.write("\n")
            file.write(",".join(map(str, vac_histogram)))
            file.write("\n")
    return isizes, vsizes


def read_count_0d(out_dir: Path) -> tuple[np.ndarray, np.ndarray]:
    """Reads interstitial and vacancy counts from a CSV file.

    Parameters
    ----------
    out_dir : Path
        Directory where the histograms were saved.

    Returns
    -------
    tuple[dtypes.INT_CHECK, dtypes.INT_CHECK]
        A tuple containing: the sizes of the interstitials and the sizes of the vacancies.
    """
    with open(out_dir / "oclusters0D.csv", "r", encoding="utf-8") as file:
        interstitials = np.array(
            list(map(int, file.readline()[:-1].split(","))), dtype=int
        )
        vacancies = np.array(list(map(int, file.readline()[:-1].split(","))), dtype=int)
    return interstitials, vacancies


# endregion



"""This module contains the `E2Recoil` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class E2Recoil(SRIMFile):
    """Class to handle `E2RECOIL.txt` file."""

    def __init__(self, srimdb: "SRIMDB") -> None:
        """Initializes the `E2Recoil` object.

        Parameters
        ----------
        srimdb : SRIMDB
            `SRIMDB` object.
        """
        super().__init__(srimdb)
        if self.srim.calculation == "quick":
            self.process_file = self.__process_file_qc
            self.merge = self.__merge_qc
        else:
            self.process_file = self.__process_file_fc
            self.merge = self.__merge_fc

    def __process_file_qc(self, e2recoil_path: Path) -> None:
        """Processes `E2RECOIL.txt` file in Quick-Calculation mode.

        Parameters
        ----------
        e2recoil_path : Path
            `E2RECOIL.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE e2recoil"
                "(depth REAL, energy_ions REAL, energy_absorbed REAL)"
            )
        )
        with open(e2recoil_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("   DEPTH"):
                    break
            next(file)
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                energy_ions = data[1]
                energy_absorbed = data[2]
                cur.execute(
                    "INSERT INTO e2recoil(depth, energy_ions, energy_absorbed) VALUES(?, ?, ?)",
                    [depth, energy_ions, energy_absorbed],
                )
        cur.close()
        self.srim.commit()

    def __process_file_fc(self, e2recoil_path: Path) -> None:
        """Processes `E2RECOIL.txt` file in Full-Calculation mode.

        Parameters
        ----------
        e2recoil_path : Path
            `E2RECOIL.txt` path.
        """
        energy_absorbed_1 = ", ".join(
            f"energy_absorbed_{i}_{j} REAL"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        energy_absorbed_2 = ", ".join(
            f"energy_absorbed_{i}_{j}"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        energy_absorbed_3 = ", ".join(
            ["?" for _ in range(len(energy_absorbed_1.split(", ")))]
        )
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE e2recoil"
                f"(depth REAL, energy_ions REAL, {energy_absorbed_1})"
            )
        )
        with open(e2recoil_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("   DEPTH"):
                    break
            next(file)
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                energy_ions = data[1]
                energy_absorbed = data[2:]
                cur.execute(
                    (
                        f"INSERT INTO e2recoil(depth, energy_ions, {energy_absorbed_2})"
                        f"VALUES(?, ?, {energy_absorbed_3})"
                    ),
                    [depth, energy_ions, *energy_absorbed],
                )
        cur.close()
        self.srim.commit()

    def __merge_qc(self, srimdb2: "SRIMDB") -> None:
        """Merges the e2recoil table with another database for Quick-Calculation mode.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        e2recoil1 = np.array(list(self.read()))
        e2recoil2 = np.array(list(srimdb2.e2recoil.read()))
        e2recoil1[:, 1:] *= nions1
        e2recoil2[:, 1:] *= nions2
        e2recoil1[:, 1:] += e2recoil2[:, 1:]
        e2recoil1[:, 1:] /= nions
        cur = self.cursor()
        cur.execute("DELETE FROM e2recoil")
        cur.executemany(
            "INSERT INTO e2recoil(depth, energy_ions, energy_absorbed) VALUES(?, ?, ?)",
            e2recoil1,
        )
        cur.close()
        self.srim.commit()

    def __merge_fc(self, srimdb2: "SRIMDB") -> None:
        """Merges the e2recoil table with another database for Full-Calculation mode.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        e2recoil1 = np.array(list(self.read()))
        e2recoil2 = np.array(list(srimdb2.e2recoil.read()))
        e2recoil1[:, 1:] *= nions1
        e2recoil2[:, 1:] *= nions2
        e2recoil1[:, 1:] += e2recoil2[:, 1:]
        e2recoil1[:, 1:] /= nions
        energy_absorbed_1 = ", ".join(
            f"energy_absorbed_{i}_{j} REAL"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        energy_absorbed_2 = ", ".join(
            f"energy_absorbed_{i}_{j}"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        energy_absorbed_3 = ", ".join(
            ["?" for _ in range(len(energy_absorbed_1.split(", ")))]
        )
        cur = self.cursor()
        cur.execute("DELETE FROM e2recoil")
        cur.executemany(
            (
                f"INSERT INTO e2recoil(depth, energy_ions, {energy_absorbed_2})"
                f"VALUES(?, ?, {energy_absorbed_3})"
            ),
            e2recoil1,
        )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads e2recoil data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM e2recoil {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()



"""This module provides a class to find and analyze defects in crystal structures."""

from dataclasses import dataclass
from typing import Optional, Union

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation as R

from irradiapy import dtypes


@dataclass
class DefectsFinder:
    """Class to find and analyze defects in crystal structures.

    Note
    ----
    Not ready to be used. Targeted to LAMMPS dump results.

    Note
    ----
    Only for bcc lattice.

    Note
    ----
    Even if an atom is very far away from the simulation box, the closest lattice
    site will be asigned to it. This is a limitation of the current implementation.
    A cutoff distance should be implemented to avoid this and consider the atom as
    free.

    Note
    ----
    Only one unit cell around the simulation box is considered for periodic
    boundary conditions.
    """

    lattice: str
    a0: float
    perx: bool
    pery: bool
    perz: bool
    nxhi: int
    nyhi: int
    nzhi: int
    nxlo: int = 0
    nylo: int = 0
    nzlo: int = 0

    FinderAtom = np.dtype(
        [("type", int), ("pos", float, 3), ("index", int, 4), ("modular", float, 3)]
    )
    FinderAtom_Check = np.ndarray[tuple[int, ...], FinderAtom]
    UnitCell = np.dtype([("pos", float, 3), ("index", int, 4)])
    UnitCell_Check = np.ndarray[tuple[int, ...], UnitCell]

    def __post_init__(self) -> None:
        if self.lattice == "bcc":
            self.unit_cell = np.array(
                [
                    ([0.0, 0.0, 0.0], [0, 0, 0, 0]),
                    ([1.0, 0.0, 0.0], [1, 0, 0, 0]),
                    ([1.0, 1.0, 0.0], [1, 1, 0, 0]),
                    ([0.0, 1.0, 0.0], [0, 1, 0, 0]),
                    ([0.0, 0.0, 1.0], [0, 0, 1, 0]),
                    ([1.0, 0.0, 1.0], [1, 0, 1, 0]),
                    ([1.0, 1.0, 1.0], [1, 1, 1, 0]),
                    ([0.0, 1.0, 1.0], [0, 1, 1, 0]),
                    ([0.5, 0.5, 0.5], [0, 0, 0, 1]),
                ],
                dtype=self.UnitCell,
            )
            self.unit_cell["pos"] *= self.a0
        else:
            raise ValueError("Only bcc lattice is supported.")

        self.nx = self.nxhi - self.nxlo
        self.ny = self.nyhi - self.nylo
        self.nz = self.nzhi - self.nzlo

    def rescale_translate_rotate(
        self,
        pos: npt.NDArray[np.float64],
        a1: Union[float, np.number],
        pka_pos: npt.NDArray[np.float64],
        pka_theta: float,
        pka_phi: float,
    ) -> None:
        """Rescales, translates, and rotates the positions of atoms.

        Parameters
        ----------
        atoms : dict
            Dictionary containing atomic positions under the key "pos".
        pka_pos : array-like
            Position vector of the primary knock-on atom (PKA).
        pka_theta : float
            Polar angle (in radians) for the PKA direction.
        pka_phi : float
            Azimuthal angle (in radians) for the PKA direction.
        a0 : float or array-like
            Initial scaling factor or vector.
        a1 : float or array-like
            Final scaling factor or vector.
        """
        # Translate
        pos -= pka_pos
        # Rotation matrix, align with PKA initial direction
        xaxis = np.array([1.0, 0.0, 0.0])
        pka_dir = np.array(
            [
                np.sin(pka_theta) * np.cos(pka_phi),
                np.sin(pka_theta) * np.sin(pka_phi),
                np.cos(pka_theta),
            ]
        )
        transform = R.align_vectors([xaxis], [pka_dir])[0].as_matrix()
        # Scaling matrix
        scaling = a1 / self.a0
        scaling_matrix = (
            np.diag([scaling] * 3) if isinstance(scaling, float) else np.diag(scaling)
        )
        # Scale-rotate matrix
        transform = transform @ scaling_matrix
        # Apply scaling and rotation
        pos[:] = pos @ transform.T

    def __generate_assignments(self) -> dict:
        """Generates the corresponding lattice in index coordinates (bcc lattice).

        Returns
        -------
        dict
            Dictionary with lattice positions and assigned atoms.

        Examples
        --------
        assignments = {
            (1, 1, 1, 1): [123, 23],
            ...
        }
        lattice position with indices (1, 1, 1, 1) has two atoms assigned, whose
        numbers are 123 and 23
        """
        assignments = {
            (xi + self.nxlo, yi + self.nylo, zi + self.nzlo, sub): []
            for xi in range(self.nx)
            for yi in range(self.ny)
            for zi in range(self.nz)
            for sub in (0, 1)
        }
        return assignments

    def find(
        self,
        types: npt.NDArray[np.integer],
        xs: Optional[npt.NDArray[np.float64]] = None,
        ys: Optional[npt.NDArray[np.float64]] = None,
        zs: Optional[npt.NDArray[np.float64]] = None,
        poss: Optional[npt.NDArray[np.float64]] = None,
        a1: Optional[Union[float, np.number]] = None,
        pka_pos: Optional[npt.NDArray[np.float64]] = None,
        pka_theta: Optional[float] = None,
        pka_phi: Optional[float] = None,
        recenter: Optional[bool] = False,
    ) -> np.ndarray:
        """Finds defects in `ifile_path` and saves the info in `ofile_path`.

        Parameters
        ----------

        Returns
        -------
        dtypes.defect_check
            Array of defects.
        """
        natoms = len(types)
        atoms = np.zeros(natoms, dtype=self.FinderAtom)
        atoms["type"] = types
        if xs is None or ys is None or zs is None:
            atoms["pos"] = poss
        elif poss is None:
            poss = np.empty((len(types), 3))
            poss[:, 0] = xs
            poss[:, 1] = ys
            poss[:, 2] = zs
            atoms["pos"] = poss
        else:
            raise ValueError("Either xs, ys, zs or poss must be given.")

        # Atom (almost) index and modular coordinates
        atoms["index"][:, :-1], atoms["modular"] = np.divmod(atoms["pos"], self.a0)
        # Array of distances: each row (atom) contains the distance to each
        # unit cell position
        dist = np.sum(
            np.square(atoms["modular"][:, np.newaxis] - self.unit_cell["pos"]), axis=2
        )
        # Get index corrdinates for the atom
        atoms["index"] += self.unit_cell["index"][np.argmin(dist, axis=1)]
        # Apply periodic boundary conditions
        self.__apply_boundary_conditions(atoms)
        # Create all lattice positions and assign a lattice position to each atom
        assignments = self.__generate_assignments()
        print(f"Number of atoms: {natoms}")
        print(f"Lattice sites:   {len(assignments)}")
        for i in range(natoms):
            assignments[tuple(atoms["index"][i])].append(i)
        # Vacancies and interstitials identification
        defects = self.__defect_identification(assignments, atoms)
        # Set the origin to the PKA position and align its direction
        # with the x-axis
        if len(defects) and recenter:
            if (
                a1 is not None
                and pka_pos is not None
                and pka_theta is not None
                and pka_phi is not None
            ):
                self.rescale_translate_rotate(
                    defects["pos"], a1, pka_pos, pka_theta, pka_phi
                )
            else:
                defects["pos"] -= np.mean(defects["pos"], axis=0)
                if a1 is not None:
                    defects["pos"] *= a1
        nvacs = len(np.where(defects["type"] == 0)[0])
        nsias = len(defects) - nvacs
        print(f"Number of interstitials: {nsias}")
        print(f"Number of vacancies: {nvacs}")
        return defects

    def __apply_boundary_conditions(
        self, atoms: "DefectsFinder.FinderAtom_Check"
    ) -> None:
        """Applies boundary conditions to the atomic index coordinates.

        Note
        ----
        Only one unit cell around the simulation box is considered for periodic
        boundary conditions.

        Parameters
        ----------
        atoms : npt.NDArray[np.float64]
            Atomic positions.
        """
        if self.perx:
            atoms["index"][:, 0][atoms["index"][:, 0] == self.nxhi] = self.nxlo
            atoms["index"][:, 0][atoms["index"][:, 0] == self.nxlo - 1] = self.nxhi - 1
        if self.pery:
            atoms["index"][:, 1][atoms["index"][:, 1] == self.nyhi] = self.nylo
            atoms["index"][:, 1][atoms["index"][:, 1] == self.nylo - 1] = self.nyhi - 1
        if self.perz:
            atoms["index"][:, 2][atoms["index"][:, 2] == self.nzhi] = self.nzlo
            atoms["index"][:, 2][atoms["index"][:, 2] == self.nzlo - 1] = self.nzhi - 1

    def __index_to_normalised_cartesian(
        self, index: tuple[int, int, int, int]
    ) -> npt.NDArray[np.float64]:
        """Converts index coordinates to normalised Cartesian coordinates.

        Parameters
        ----------
        index : tuple
            Index coordinates.

        Returns
        -------
        npt.NDArray[np.float64]
            Normalised Cartesian coordinates.
        """
        return (np.array(index[:-1]) + index[-1] * 0.5) * self.a0

    def __defect_identification(
        self, assignments: dict, atoms: "DefectsFinder.FinderAtom_Check"
    ) -> np.ndarray:
        defects = np.empty(0, dtypes.defect)
        for key, value in assignments.items():
            # lattice site cartesian coordinates
            site_pos = self.__index_to_normalised_cartesian(key)
            # no atoms assigned > vacancy
            if len(value) == 0:
                defects = np.concatenate(
                    (defects, np.array([(0, site_pos)], dtype=dtypes.defect))
                )
            elif len(value) > 1:
                # do not consider as defect the closest atom to the lattice site
                dist = np.sum(np.square(site_pos - atoms[value]["pos"]))
                mask = np.full(len(value), True)
                mask[np.argmin(dist)] = False
                inters = np.empty(len(value) - 1, dtypes.defect)
                inters["type"] = atoms[value]["type"][mask]
                inters["pos"] = atoms[value]["pos"][mask]
                defects = np.concatenate((defects, inters))
        return defects



"""Utility functions related to collisions' debris from SRIM data."""

import pathlib
from typing import Optional, Union

import numpy as np

from irradiapy import dpa, dtypes, materials
from irradiapy.damagedb import DamageDB
from irradiapy.io.xyzwriter import XYZWriter
from irradiapy.srimpy.srimdb import SRIMDB


def generate_debris(
    srimdb: SRIMDB,
    dir_mddb: pathlib.Path,
    compute_tdam: bool,
    path_collisions: pathlib.Path,
    dpa_mode: dpa.DpaMode,
    add_injected: bool,
    outsiders: bool,
    seed: Optional[int] = None,
    depth_offset: Union[int, float, np.number] = 0.0,
) -> None:
    """Turns SRIM's collisions into `.xyz` files for the given database of cascades' debris.

    Warning
    -------
        Assumes a monolayer monoatomic target and same element for all ions.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    dir_mddb : pathlib.Path
        Directory where the database of cascades' debris is stored.
    compute_tdam : bool
        Whether to transform the PKA energies into damage energies. It should be `True` for
        MD simulations without electronic stopping.
    path_collisions : pathlib.Path
        Directory where the ions debris will be stored as `.xyz` files.
    dpa_mode : dpa.DpaMode
        Formula to convert the residual energy into Frenkel pairs.
    add_injected : bool
        Whether to add the injected interstitial.
    outsiders : bool
        Whether to include defects generated outside the material (PKAs close to surfaces).
    seed : int, optional
        Random seed for reproducibility.
    depth_offset : float, optional
        Offset to add to the depth of the defects, by default 0.0.
    """
    depth_min, depth_max = 0.0, srimdb.target.layers[0].width
    mat_pka = materials.get_material_by_atomic_number(
        next(
            iter(srimdb.trimdat.read(what="atom_numb", condition="WHERE ion_numb = 1"))
        )[0]
    )
    mat_target = materials.get_material_by_atomic_number(
        srimdb.target.layers[0].elements[0].atomic_number
    )
    nions = srimdb.nions
    damagedb = DamageDB(
        dir_mddb=dir_mddb,
        compute_tdam=compute_tdam,
        mat_pka=mat_pka,
        mat_target=mat_target,
        dpa_mode=dpa_mode,
        seed=seed,
    )
    with XYZWriter(path_collisions) as writer:
        for nion in range(1, nions + 1):
            writer.save(
                __generate_ion_defects(
                    srimdb,
                    nion,
                    damagedb,
                    add_injected,
                    depth_min,
                    depth_max,
                    outsiders,
                    depth_offset,
                ),
                extra_comment=f"Projectile={nion}",
            )


def __generate_ion_defects(
    srimdb: SRIMDB,
    nion: int,
    damagedb: DamageDB,
    add_injected: bool,
    depth_min: float,
    depth_max: float,
    outsiders: bool,
    depth_offset: Union[int, float, np.number] = 0.0,
) -> np.ndarray:
    """Generates the defects for a specific ion in the SRIM simulation.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    nion : int
        Ion number.
    damagedb : DamageDB
        DamageDB class that will choose MD debris.
    add_injected : bool
        Whether to add the injected interstitial.
    depth_min : float
        Minimum depth to consider for defects.
    depth_max : float
        Maximum depth to consider for defects.
    outsiders : bool
        Whether to include defects generated outside the material (PKAs close to surface).
    depth_offset : float, optional
        Offset to add to the depth of the defects, by default 0.0.

    Returns
    -------
    np.ndarray
        An array containing the defects generated by a single ion.
    """
    defects = np.empty(0, dtype=dtypes.defect)
    for depth, y, z, cosx, cosy, cosz, pka_e in srimdb.collision.read(
        what="depth, y, z, cosx, cosy, cosz, recoil_energy",
        condition=f"WHERE ion_numb = {nion}",
    ):
        pka_pos = np.array([depth, y, z])
        pka_dir = np.array([cosx, cosy, cosz])
        defects_ = damagedb.get_pka_debris(
            pka_e=pka_e, pka_pos=pka_pos, pka_dir=pka_dir
        )
        defects = np.concatenate((defects, defects_))

    if add_injected:
        injected = list(
            srimdb.range3d.read(
                what="depth, y, z", condition=f"WHERE ion_numb = {nion} LIMIT 1"
            )
        )
        if injected:
            atom_number = list(
                srimdb.trimdat.read(
                    what="atom_numb", condition=f"WHERE ion_numb = {nion}"
                )
            )[0][0]
            injected = np.array([(atom_number, injected[0])], dtype=dtypes.defect)
            defects = np.concatenate((defects, injected))
    if not outsiders:
        defects = defects[
            (defects["pos"][:, 0] >= depth_min) & (defects["pos"][:, 0] <= depth_max)
        ]
    defects["pos"][:, 0] += depth_offset
    return defects



"""This subpackage provides classes for reading and writing various file formats."""

from .lammpsreader import LAMMPSReader
from .lammpswriter import LAMMPSWriter
from .xyzreader import XYZReader
from .xyzwriter import XYZWriter



"""This module contains the `XYZReader` class."""

import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generator

import numpy as np
import numpy.typing as npt


@dataclass
class XYZReader:
    """A class to read data from an extended XYZ file.

    Attributes
    ----------
    file_path : Path
        The path to the XYZ file.
    dtype : npt.DTypeLike
        The data type of the properties in the file.
    """

    file_path: Path
    dtype: npt.DTypeLike = field(default=None, init=False)

    def _get_properties(
        self, comment: str
    ) -> tuple[list[str], list[type], list[int], np.dtype]:
        """Sets properties using the comment line.

        Parameters
        ----------
        comment : str
            Comment line.

        Returns
        -------
        tuple[list[str], list[type], list[int], np.dtype]
            Properties names, types, multiplicities, and dtype.
        """
        match = re.search(r"Properties=([^ \n]+)", comment)
        if not match:
            raise ValueError("Missing or invalid comment line format.")
        properties = match.group(1).split(":")
        num_properties = len(properties) // 3
        name_props = [properties[i * 3] for i in range(num_properties)]
        type_props = [
            self._map_type(properties[i * 3 + 1]) for i in range(num_properties)
        ]
        multiplicity_props = [int(properties[i * 3 + 2]) for i in range(num_properties)]
        dtype = np.dtype(
            [
                (
                    (name_props[i], type_props[i])
                    if multiplicity_props[i] == 1
                    else (
                        name_props[i],
                        type_props[i],
                        multiplicity_props[i],
                    )
                )
                for i in range(num_properties)
            ]
        )
        return name_props, type_props, multiplicity_props, dtype

    @staticmethod
    def _get_kind(kind: str) -> type:
        """Maps dtype kind to Python type.

        Parameters
        ----------
        kind : str
            Kind of the dtype.
        """
        if kind == "i":
            return int
        elif kind == "f":
            return float
        elif kind == "U":
            return str
        else:
            raise TypeError("Unexpected dtype")

    @staticmethod
    def _map_type(type_str: str) -> type:
        """Maps type string to Python type.

        Parameters
        ----------
        type_str : str
            Type string.
        """
        if type_str == "S":
            return str
        elif type_str == "I":
            return int
        elif type_str == "R":
            return float
        else:
            raise TypeError(f"Unexpected type string: {type_str}")

    def _line_to_data(
        self,
        line: str,
        name_props: list[str],
        type_props: list[type],
        multiplicity_props: list[int],
        dtype: np.dtype,
    ) -> npt.ArrayLike:
        """Turns one line of data into a numpy array.

        Parameters
        ----------
        line : str
            Line containing the data.
        name_props : list[str]
            Names of the properties.
        type_props : list[type]
            Types of the properties.
        multiplicity_props : list[int]
            Multiplicities of the properties.
        dtype : np.dtype
            Data type of the properties.

        Returns
        -------
        npt.ArrayLike
            The data in the line.
        """
        output = np.empty(1, dtype=dtype)
        data = line.split()
        col = 0
        for i, name_prop in enumerate(name_props):
            multiplicity_prop = multiplicity_props[i]
            type_prop = type_props[i]
            if multiplicity_prop == 1:
                output[name_prop] = type_prop(data[col])
            else:
                output[name_prop] = [
                    type_prop(data[col + j]) for j in range(multiplicity_prop)
                ]
            col += multiplicity_prop
        return output[0]

    def __iter__(self) -> Generator[Any, None, None]:
        """Iterate over subfiles in the XYZ file.

        Yields
        ------
        np.ndarray
            Array of atom data for each subfile.
        """
        with self.file_path.open(encoding="utf-8") as file:
            while True:
                line = file.readline()
                if not line:
                    break
                natoms = int(line)
                line = file.readline()
                name_props, type_props, multiplicity_props, dtype = (
                    self._get_properties(line)
                )
                atoms = np.empty(natoms, dtype=dtype)
                for i in range(natoms):
                    line = file.readline()
                    atoms[i] = self._line_to_data(
                        line, name_props, type_props, multiplicity_props, dtype
                    )
                yield atoms



"""This module contains the `Vacancy` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Vacancy(SRIMFile):
    """Class to handle `VACANCY.txt` file."""

    def __init__(self, srimdb: "SRIMDB") -> None:
        """Initializes the `Vacancy` object.

        Parameters
        ----------
        srimdb : SRIMDB
            `SRIMDB` object.
        """
        super().__init__(srimdb)
        if self.srim.calculation == "quick":
            self.process_file = self.__process_file_qc
            self.merge = self.__merge_qc
        else:
            self.process_file = self.__process_file_fc
            self.merge = self.__merge_fc

    def __process_file_qc(self, vacancy_path: Path) -> None:
        """Processes `VACANCY.txt` file in Quick-Calculation mode.

        Parameters
        ----------
        vacancy_path : Path
            `VACANCY.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE vacancy"
                "(depth REAL, vacancies_ions REAL, vacancies_recoils REAL)"
            )
        )
        with open(vacancy_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("   TARGET"):
                    break
            next(file)
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                vacancies_ions = data[1]
                vacancies_recoils = data[2]
                cur.execute(
                    (
                        "INSERT INTO vacancy"
                        "(depth, vacancies_ions, vacancies_recoils)"
                        "VALUES(?, ?, ?)"
                    ),
                    [depth, vacancies_ions, vacancies_recoils],
                )
        cur.close()
        self.srim.commit()

    def __process_file_fc(self, vacancy_path: Path) -> None:
        """Processes `VACANCY.txt` file in Full-Calculation mode.

        Parameters
        ----------
        vacancy_path : Path
            `VACANCY.txt` path.
        """
        vacancies_1 = ", ".join(
            f"vacancies_{i}_{j} REAL"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        vacancies_2 = ", ".join(
            f"vacancies_{i}_{j}"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        vacancies_3 = ", ".join(["?" for _ in range(len(vacancies_1.split(", ")))])
        cur = self.cursor()
        cur.execute(
            ("CREATE TABLE vacancy" f"(depth REAL, knock_ons REAL, {vacancies_1})")
        )
        with open(vacancy_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("   TARGET"):
                    break
            next(file)
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                knock_ons = data[1]
                vacancies = data[2:]
                cur.execute(
                    (
                        "INSERT INTO vacancy"
                        f"(depth, knock_ons, {vacancies_2})"
                        f"VALUES(?, ?, {vacancies_3})"
                    ),
                    [depth, knock_ons, *vacancies],
                )
        cur.close()
        self.srim.commit()

    def __merge_qc(self, srimdb2: "SRIMDB") -> None:
        """Merges the e2recoil table with another database for Quick-Calculation mode.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge with.
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        vacancy1 = np.array(list(self.read()))
        vacancy2 = np.array(list(srimdb2.vacancy.read()))
        vacancy1[:, 1:] *= nions1
        vacancy2[:, 1:] *= nions2
        vacancy1[:, 1:] += vacancy2[:, 1:]
        vacancy1[:, 1:] /= nions
        cur = self.cursor()
        cur.execute("DELETE FROM vacancy")
        cur.executemany(
            (
                "INSERT INTO vacancy(depth, vacancies_ions, vacancies_recoils)"
                "VALUES(?, ?, ?)"
            ),
            vacancy1,
        )
        cur.close()
        self.srim.commit()

    def __merge_fc(self, srimdb2: "SRIMDB") -> None:
        """Merges the vacancy table with another database for Full-Calculation mode.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge with
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        vacancy1 = np.array(list(self.read()))
        vacancy2 = np.array(list(srimdb2.vacancy.read()))
        vacancy1[:, 1:] *= nions1
        vacancy2[:, 1:] *= nions2
        vacancy1[:, 1:] += vacancy2[:, 1:]
        vacancy1[:, 1:] /= nions
        vacancies_1 = ", ".join(
            f"vacancies_{i}_{j} REAL"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        vacancies_2 = ", ".join(
            f"vacancies_{i}_{j}"
            for i, layer in enumerate(self.srim.target.layers)
            for j in range(len(layer.elements))
        )
        vacancies_3 = ", ".join(["?" for _ in range(len(vacancies_1.split(", ")))])
        cur = self.cursor()
        cur.execute("DELETE FROM vacancy")
        cur.executemany(
            (
                f"INSERT INTO vacancy(depth, knock_ons, {vacancies_2})"
                f"VALUES(?, ?, {vacancies_3})"
            ),
            vacancy1,
        )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads vacancy data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM vacancy {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()



"""This module contains the `Range3D` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Range3D(SRIMFile):
    """Class to handle `RANGE_3D.txt` file."""

    def process_file(self, range3d_path: Path) -> None:
        """Processes `RANGE_3D.txt` file.

        Parameters
        ----------
        range3d_path : Path
            `RANGE_3D.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            "CREATE TABLE range3d(ion_numb INTEGER, depth REAL, y REAL, z REAL)"
        )
        with open(range3d_path, "r", encoding="latin1") as file:
            for line in file:
                if line.startswith("Number"):
                    break
            next(file)
            for line in file:
                data = list(map(float, line[:-1].split()))
                ion_numb = data[0]
                depth = data[1]
                y = data[2]
                z = data[3]
                cur.execute(
                    "INSERT INTO range3d(ion_numb, depth, y, z) VALUES(?, ?, ?, ?)",
                    [ion_numb, depth, y, z],
                )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads range3d data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM range3d {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the range3d table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions = self.srim.nions
        cur = self.cursor()
        cur.execute(f"ATTACH DATABASE '{srimdb2.db_path}' AS srimdb2")
        cur.execute(
            (
                "INSERT INTO range3d(ion_numb, depth, y, z)"
                "SELECT ion_numb + ?, depth, y, z FROM srimdb2.range3d"
            ),
            (nions,),
        )
        self.srim.commit()
        cur.execute("DETACH DATABASE srimdb2")
        cur.close()



"""This module contains the `SRIMFile` class."""

import sqlite3
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


@dataclass
class SRIMFile:
    """Base class for SRIM files."""

    srim: "SRIMDB"

    def cursor(self) -> sqlite3.Cursor:
        """Returns a cursor object using the current SRIM database connection.

        Returns
        -------
        sqlite3.Cursor
            A cursor object for the SRIM database connection.
        """
        return self.srim.cursor()

    def commit(self) -> None:
        """Commits the current transaction."""
        self.srim.commit()



"""This module contains the `Transmit` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Transmit(SRIMFile):
    """Class to handle `TRANSMIT.txt` file."""

    def process_file(self, transmit_path: Path) -> None:
        """Processes `TRANSMIT.txt` file.

        Parameters
        ----------
        transmit_path : Path
            `TRANSMIT.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE transmit"
                "(ion_numb INTEGER, atom_numb INTEGER, energy REAL, depth REAL,"
                "y REAL, z REAL, cosx REAL, cosy REAL, cosz REAL)"
            )
        )
        with open(transmit_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith(" Numb"):
                    break
            cur = self.cursor()
            for line in file:
                line = line[1:-2]
                data = list(map(float, line[:-1].split()))
                ion_numb = data[0]
                atom_numb = data[1]
                energy = data[2]
                depth = data[3]
                y = data[4]
                z = data[5]
                cosx = data[6]
                cosy = data[7]
                cosz = data[8]
                cur.execute(
                    (
                        "INSERT INTO transmit"
                        "(ion_numb, atom_numb, energy, depth, y, z, cosx, cosy, cosz)"
                        "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?)"
                    ),
                    [ion_numb, atom_numb, energy, depth, y, z, cosx, cosy, cosz],
                )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads transmit data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM transmit {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def get_nions(self) -> int:
        """Returns the number of ions in the database.

        Returns
        -------
        int
            Number of ions.
        """
        cur = self.cursor()
        cur.execute("SELECT COUNT(1) FROM transmit")
        nions = cur.fetchone()[0]
        cur.close()
        return nions

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the transmit table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions = self.srim.transmit.get_nions()
        cur = self.cursor()
        cur.execute(f"ATTACH DATABASE '{srimdb2.db_path}' AS srimdb2")
        cur.execute(
            (
                "INSERT INTO transmit"
                "(ion_numb, atom_numb, energy, depth, y, z, cosx, cosy, cosz)"
                "SELECT ion_numb + ?, atom_numb, energy, depth, y, z, cosx, "
                "cosy, cosz FROM srimdb2.transmit"
            ),
            (nions,),
        )
        self.srim.commit()
        cur.execute("DETACH DATABASE srimdb2")
        cur.close()



"""Subpackage for SRIM output files."""

from .backscat import Backscat
from .collision import Collision
from .e2recoil import E2Recoil
from .ioniz import Ioniz
from .lateral import Lateral
from .novac import Novac
from .phonon import Phonon
from .range import Range
from .range3d import Range3D
from .sputter import Sputter
from .srimfile import SRIMFile
from .subcollision import Subcollision
from .transmit import Transmit
from .trimdat import Trimdat
from .vacancy import Vacancy



"""Contains definitions and utilities related to dpa."""

from enum import Enum, auto
from typing import Union

import numpy as np

from irradiapy import materials


class DpaMode(Enum):
    """Enumeration of dpa calculation modes.

    References
    ----------
    NRT : https://doi.org/10.1016/0029-5493(75)90035-7
    ARC : https://doi.org/10.1038/s41467-018-03415-5
    FERARC : https://doi.org/10.1103/PhysRevMaterials.5.073602
    """

    NRT = auto()
    ARC = auto()
    FERARC = auto()


def compute_damage_energy(
    epka: Union[float, np.ndarray],
    mat_pka: materials.Material,
    mat_target: materials.Material,
    force_lss: bool = False,
) -> Union[float, np.ndarray]:
    """Compute damage energy from the given PKA energy.

    Uses SRIM stopping powers if available and `force_lss` is False.
    Otherwise uses the Lindhard formula.

    Parameters
    ----------
    epka : float or np.ndarray
        PKA energy in electron volts.
    mat_pka : materials.Material
        Material of the primary knock-on atom.
    mat_target : materials.Material
        Target material in which the knock-on occurs.
    force_lss : bool, optional
        If True, force the use of the Lindhard formula. Default is False.

    Returns
    -------
    float or np.ndarray
        Damage energy in electron volts.
    """
    fe = materials.get_material_by_name("Fe")
    w = materials.get_material_by_name("W")
    if not force_lss:
        if mat_pka == fe and mat_target == fe:
            #  SRIM Quick-Calculation, D1
            return 699e-3 * epka - 460e-9 * np.square(epka)
        if mat_pka == w and mat_target == w:
            # SRIM Quick-Calculation, D1
            return 752e-3 * epka - 216e-9 * np.square(epka)
    # Otherwise use Lindhard formula
    a0 = 0.529177e-10  # m, Bohr radius
    e2 = 1.4e-9  # eV2 m s, squared unit charge for Lindhard expression
    a = (
        (9.0 * np.pi**2 / 128.0) ** (1.0 / 3.0)
        * a0
        / (
            mat_pka.atomic_number ** (2.0 / 3.0)
            + mat_target.atomic_number ** (2.0 / 3.0)
        )
        ** 0.5
    )
    redu = (
        (mat_target.mass_number * epka)
        / (mat_pka.mass_number + mat_target.mass_number)
        * a
        / (mat_pka.atomic_number * mat_target.atomic_number * e2)
    )
    k = (
        0.1337
        * mat_pka.atomic_number ** (1.0 / 6.0)
        * (mat_pka.atomic_number / mat_pka.mass_number) ** 0.5
    )
    g = 3.4008 * redu ** (1.0 / 6.0) + 0.40244 * redu ** (3.0 / 4.0) + redu
    return epka / (1.0 + k * g)


def calc_nrt_dpa(
    tdam: Union[int, float, np.ndarray],
    mat: materials.Material,
) -> Union[int, np.ndarray]:
    """Calculate the NRT-dpa for the given damage energy.

    Parameters
    ----------
    tdam : int, float, or numpy.ndarray
        Damage energy in electron volts.
    mat : materials.Material
        Material.

    Returns
    -------
    int or numpy.ndarray
        Number of Frenkel pairs predicted by NRT-dpa.
    """
    min_threshold = mat.ed_avr
    max_threshold = 2.5 * mat.ed_avr

    def scaling_func(x):
        return 0.4 * x / mat.ed_avr

    if isinstance(tdam, (float, int)):
        if tdam < min_threshold:
            return 0.0
        elif tdam > max_threshold:
            return scaling_func(tdam)
        else:
            return 1.0
    elif isinstance(tdam, np.ndarray) and np.issubdtype(tdam.dtype, np.number):
        return _apply_dpa_thresholds(tdam, min_threshold, max_threshold, scaling_func)
    else:
        raise TypeError("tdam must be a number or numpy array of numbers")


def calc_arc_dpa(
    tdam: Union[int, float, np.ndarray],
    mat: materials.Material,
) -> Union[int, np.ndarray]:
    """Calculate the arc-dpa for the given damage energy in eV.

    Parameters
    ----------
    tdam : int, float, or numpy.ndarray
        Damage energy in electron volts.
    mat : materials.Material
        Material in which damage is calculated.

    Returns
    -------
    int or numpy.ndarray
        Number of Frenkel pairs predicted by arc-dpa.
    """
    min_threshold = mat.ed_avr
    max_threshold = 2.5 * mat.ed_avr

    def scaling_func(x):
        return 0.4 * x / mat.ed_avr

    def efficiency_func(x):
        return (1.0 - mat.c_arc) / (max_threshold**mat.b_arc) * x**mat.b_arc + mat.c_arc

    if isinstance(tdam, (float, int)):
        if tdam < min_threshold:
            return 0.0
        elif tdam > max_threshold:
            eff = efficiency_func(tdam)
            return scaling_func(tdam) * eff
        else:
            return 1.0
    elif isinstance(tdam, np.ndarray):
        return _apply_dpa_thresholds(
            tdam, min_threshold, max_threshold, scaling_func, efficiency_func
        )
    else:
        raise TypeError("tdam must be a number or numpy array of numbers")


def calc_fer_arc_dpa(
    tdam: Union[int, float, np.ndarray],
    mat: materials.Material,
) -> Union[int, np.ndarray]:
    """Calculate the fer-arc-dpa for the given damage energy.

    Parameters
    ----------
    tdam : int, float, or numpy.ndarray
        Damage energy in electron volts.
    mat : materials.Material
        Material.

    Returns
    -------
    int or numpy.ndarray
        Number of Frenkel pairs predicted by modified arc-dpa.
    """
    min_threshold = mat.ed_min
    max_threshold = 2.5 * mat.ed_avr

    def scaling_func(x):
        return 0.4 * x / mat.ed_avr

    def efficiency_func(x):
        return (1.0 - mat.c_arc) / (max_threshold**mat.b_arc) * x**mat.b_arc + mat.c_arc

    if isinstance(tdam, (float, int)):
        if tdam < min_threshold:
            return 0.0
        elif tdam > max_threshold:
            eff = efficiency_func(tdam)
            return scaling_func(tdam) * eff
        else:
            return scaling_func(tdam)
    elif isinstance(tdam, np.ndarray):
        return _apply_dpa_thresholds(
            tdam,
            min_threshold,
            max_threshold,
            scaling_func,
            efficiency_func,
            scaling_func,
        )
    else:
        raise TypeError("tdam must be a number or numpy array of numbers")


def _apply_dpa_thresholds(
    tdam: np.ndarray,
    min_threshold: float,
    max_threshold: float,
    scaling_func: callable,
    efficiency_func: callable = None,
    middle_func: callable = None,
):
    """Apply dpa thresholds and scaling/efficiency functions.

    Parameters
    ----------
    tdam : np.ndarray
        Damage energy array.
    min_threshold : float
        Minimum threshold for dpa.
    max_threshold : float
        Maximum threshold for dpa.
    scaling_func : callable
        Function to scale damage energy.
    efficiency_func : callable, optional
        Efficiency function for high energies. Default is None.
    middle_func : callable, optional
        Function for values between thresholds. Default is None.

    Returns
    -------
    np.ndarray
        Array of dpa values.
    """
    result = np.ones_like(tdam, dtype=np.float64)
    below_mask = tdam < min_threshold
    above_mask = tdam > max_threshold
    result[below_mask] = 0
    if middle_func is not None:
        middle_mask = (~below_mask) & (~above_mask)
        result[middle_mask] = middle_func(tdam[middle_mask])
    # else: keep as 1
    if efficiency_func:
        result[above_mask] = scaling_func(tdam[above_mask]) * efficiency_func(
            tdam[above_mask]
        )
    else:
        result[above_mask] = scaling_func(tdam[above_mask])
    return result



"""This module contains the `Ioniz` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Lateral(SRIMFile):
    """Class to handle `LATERAL.txt` file."""

    def process_file(self, lateral_path: Path) -> None:
        """Processes `LATERAL.txt` file.

        Parameters
        ----------
        lateral_path : Path
            `LATERAL.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE lateral"
                "(depth REAL, lateral_proj_range REAL,"
                "projected_straggling REAL, lateral_radial REAL,"
                "radial_straggling REAL)"
            )
        )
        with open(lateral_path, "r", encoding="latin1") as file:
            for line in file:
                if line.startswith("  TARGET"):
                    break
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                lateral_proj_range = data[1]
                projected_straggling = data[2]
                lateral_radial = data[3]
                radial_straggling = data[4]
                cur.execute(
                    (
                        "INSERT INTO lateral"
                        "(depth, lateral_proj_range, projected_straggling,"
                        "lateral_radial, radial_straggling)"
                        "VALUES(?, ?, ?, ?, ?)"
                    ),
                    [
                        depth,
                        lateral_proj_range,
                        projected_straggling,
                        lateral_radial,
                        radial_straggling,
                    ],
                )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads lateral data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM lateral {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the lateral table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.

        Note
        ----
        Not sure if calculations are right, see Eqs. 9-1, 9-2 and 9-3,
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        lateral1 = np.array(list(self.read()))
        lateral2 = np.array(list(srimdb2.lateral.read()))
        # Lateral projected range
        lateral1[:, 1] *= nions1
        lateral2[:, 1] *= nions2
        lateral1[:, 1] += lateral2[:, 1]
        lateral1[:, 1] /= nions
        # Projected straggling
        lateral1[:, 2] **= 2.0
        lateral2[:, 2] **= 2.0
        lateral1[:, 2] *= nions1
        lateral2[:, 2] *= nions2
        lateral1[:, 2] += lateral2[:, 2]
        lateral1[:, 2] /= nions
        lateral1[:, 2] **= 0.5
        # Lateral radial
        lateral1[:, 3] *= nions1
        lateral2[:, 3] *= nions2
        lateral1[:, 3] += lateral2[:, 3]
        lateral1[:, 3] /= nions
        # Radial straggling
        lateral1[:, 4] **= 2.0
        lateral2[:, 4] **= 2.0
        lateral1[:, 4] *= nions1
        lateral2[:, 4] *= nions2
        lateral1[:, 4] += lateral2[:, 4]
        lateral1[:, 4] /= nions
        lateral1[:, 4] **= 0.5
        cur = self.cursor()
        cur.execute("DELETE FROM lateral")
        cur.executemany(
            "INSERT INTO lateral"
            "(depth, lateral_proj_range, projected_straggling,"
            "lateral_radial, radial_straggling)"
            "VALUES(?, ?, ?, ?, ?)",
            lateral1,
        )
        cur.close()
        self.srim.commit()



"""TThis module contains the `Target` class."""

from dataclasses import dataclass, field

from irradiapy.srimpy.target.layer import Layer


@dataclass
class Target:
    """Class for defining the target material in SRIM simulations.

    Attributes
    ----------
    layers : list
        The list of layers in the target material.
    nelements : int
        The total number of elements in the target material.
    nlayers : int
    """

    layers: list[Layer]
    nelements: int = field(init=False)
    nlayers: int = field(init=False)

    def __post_init__(self) -> None:
        """Counts the number of layers and elements in the material after initialization."""
        self.nlayers = len(self.layers)
        self.nelements = sum(len(layer.elements) for layer in self.layers)

    def trimin_description(self) -> str:
        """Return the target material description in the TRIMIN format."""
        string = (
            "target material, number of elements, layers\n"
            f'"material" {self.nelements} {self.nlayers}'
        )
        return string

    def trimin_target_elements(self) -> str:
        """Return the target elements in the TRIMIN format."""
        string = "target element, Z, mass\n"
        for i, layer in enumerate(self.layers):
            for j, element in enumerate(layer.elements):
                string += (
                    f"Atom {i*len(layer.elements)+j+1} = {element.symbol} "
                    f"=      {element.atomic_number} {element.atomic_mass}\n"
                )
        return string

    def trimin_target_layers(self) -> str:
        """Return the target layers in the TRIMIN format."""
        string = (
            "layer name, width density elements\nnumb. desc. (ang) (g/cm3) stoich.\n"
        )
        layers_info = []
        for i, layer in enumerate(self.layers):
            prev_stoichs = [0.0] * sum(len(l.elements) for l in self.layers[:i])
            next_stoichs = [0.0] * sum(len(l.elements) for l in self.layers[i + 1 :])
            layer_info = f'{i} "layer{i}" {layer.width} {layer.density} '
            layer_info += " ".join(
                map(str, prev_stoichs + layer.stoichs + next_stoichs)
            )
            layers_info.append(layer_info)
        return string + "\n".join(layers_info)

    def trimin_phases(self) -> str:
        """Return the phases of the layers in the TRIMIN format."""
        string = "phases\n"
        string += " ".join(map(str, (layer.phase for layer in self.layers)))
        return string

    def trimin_bragg(self) -> str:
        """Return the Bragg corrections in the TRIMIN format."""
        string = "target compound corrections (bragg)\n"
        string += " ".join(map(str, (layer.bragg for layer in self.layers)))
        return string

    def trimin_displacement(self) -> str:
        """Return the displacement energies in the TRIMIN format."""
        string = "displacement energies (eV)\n"
        string += " ".join(
            map(
                str,
                (element.e_d for layer in self.layers for element in layer.elements),
            )
        )
        return string

    def trimin_lattice(self) -> str:
        """Return the lattice binding energies in the TRIMIN format."""
        string = "lattice binding energies (eV)\n"
        string += " ".join(
            map(
                str,
                (element.e_l for layer in self.layers for element in layer.elements),
            )
        )
        return string

    def trimin_surface(self) -> str:
        """Return the surface binding energies in the TRIMIN format."""
        string = "surface binding energies (eV)\n"
        string += " ".join(
            map(
                str,
                (element.e_s for layer in self.layers for element in layer.elements),
            )
        )
        return string



"""This module contains the `Subcollision` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Subcollision(SRIMFile):
    """Class for processing subcollision data."""

    def __init__(self, srimdb: "SRIMDB") -> None:
        """Initializes the `Subcollision` object.

        Parameters
        ----------
        srimdb : SRIMDB
            `SRIMDB` object.
        """
        super().__init__(srimdb)
        if self.srim.calculation == "quick":
            self.process_file = self.__process_file_qc
        else:
            self.process_file = self.__process_file_fc

    def __process_file_qc(self, collision_path: Path) -> None:
        """Processes `COLLISON.txt` file in Quick-Calculation mode.

        Parameters
        ----------
        collision_path : Path
            `COLLISON.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TEMPORARY TABLE IF NOT EXISTS subcollision"
                # (f"CREATE TABLE IF NOT EXISTS subcollision"
                "(ion_numb INTEGER, energy REAL, depth REAL, y REAL, z REAL,"
                "se REAL, atom_hit TEXT, recoil_energy REAL,"
                "target_disp REAL)"
            )
        )
        with open(collision_path, "r", encoding="latin1") as file:
            for line in file:
                # Skip this line
                # ³=> Recoils Calculated with Kinchin-Pease Theory (Only Vacancies Calc) <=³
                if line[0] == "³":
                    break
            for line in file:
                if line[0] == "³":
                    line = line[1:-2]
                    data = line.split("³")
                    ion_numb = int(data[0])
                    energy = float(data[1])
                    depth = float(data[2])
                    y = float(data[3])
                    z = float(data[4])
                    se = float(data[5])
                    atom_hit = data[6].strip()
                    recoil_energies = float(data[7])
                    target_disp = float(data[8])
                    cur.execute(
                        (
                            "INSERT INTO subcollision"
                            "(ion_numb, energy, depth, y, z, se, atom_hit, recoil_energy,"
                            "target_disp)"
                            "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?)"
                        ),
                        [
                            ion_numb,
                            energy,
                            depth,
                            y,
                            z,
                            se,
                            atom_hit,
                            recoil_energies,
                            target_disp,
                        ],
                    )
        cur.close()

    def __process_file_fc(self, collision_path: Path) -> None:
        """Processes `COLLISON.txt` file in Full-Calculation mode.

        Parameters
        ----------
        collision_path : Path
            `COLLISON.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TEMPORARY TABLE IF NOT EXISTS subcollision"
                # (f"CREATE TABLE IF NOT EXISTS subcollision"
                "(ion_numb INTEGER, energy REAL, depth REAL, y REAL, z REAL,"
                "se REAL, atom_hit TEXT, recoil_energy REAL,"
                "target_disp INTEGER, target_vac INTEGER,"
                "target_replac INTEGER, target_inter INTEGER)"
            )
        )
        with open(collision_path, "r", encoding="latin1") as file:
            for line in file:
                if line[0] == "³":
                    line = line[1:-2]
                    data = line.split("³")
                    ion_numb = int(data[0])
                    energy = float(data[1])
                    depth = float(data[2])
                    y = float(data[3])
                    z = float(data[4])
                    se = float(data[5])
                    atom_hit = data[6].strip()
                    recoil_energies = float(data[7])
                    target_disp = int(data[8])
                    target_vac = int(data[9])
                    target_replac = int(data[10])
                    target_inter = int(data[11])
                    cur.execute(
                        (
                            "INSERT INTO subcollision"
                            "(ion_numb, energy, depth, y, z, se, atom_hit, recoil_energy,"
                            "target_disp, target_vac, target_replac, target_inter)"
                            "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
                        ),
                        [
                            ion_numb,
                            energy,
                            depth,
                            y,
                            z,
                            se,
                            atom_hit,
                            recoil_energies,
                            target_disp,
                            target_vac,
                            target_replac,
                            target_inter,
                        ],
                    )
        cur.close()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads data from the subcollision table.

        Parameters
        ----------
        what : str, optional
            Columns to read, by default "*".
        condition : str, optional
            Condition to filter data, by default "".

        Yields
        ------
        tuple
            Data from the subcollision table.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM subcollision {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def empty(self) -> None:
        """Empties the subcollision table."""
        cur = self.cursor()
        cur.execute("DELETE FROM subcollision")
        cur.close()



"""This module contains the `Novac` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Novac(SRIMFile):
    """Class for processing the `NOVAC.txt` file."""

    def process_file(self, novac_path: Path) -> None:
        """Processes `NOVAC.txt` file.

        Parameters
        ----------
        novac_path : Path
            `NOVAC.txt` path.
        """
        cur = self.cursor()
        cur.execute("CREATE TABLE novac(depth REAL, number REAL)")
        with open(novac_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("  DEPTH"):
                    break
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                energy_ions = data[1]
                cur.execute(
                    "INSERT INTO novac(depth, number) VALUES(?, ?)",
                    [depth, energy_ions],
                )
        cur.close()
        self.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads the novac table.

        Parameters
        ----------
        what : str
            Columns to read.
        condition : str
            Condition to filter the data.

        Yields
        ------
        tuple
            Data from the novac table.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM novac {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the novac table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        novac1 = np.array(list(self.read()))
        novac2 = np.array(list(srimdb2.novac.read()))
        novac1[:, 1:] *= nions1
        novac2[:, 1:] *= nions2
        novac1[:, 1:] = novac1[:, 1:] + novac2[:, 1:]
        novac1[:, 1:] /= nions1 + nions2
        cur = self.cursor()
        cur.execute("DELETE FROM novac")
        cur.executemany("INSERT INTO novac(depth, number) VALUES(?, ?)", novac1)
        cur.close()
        self.commit()



"""This module provides a class to find and analyze defects in crystal structures."""

import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation as R

from irradiapy import dtypes
from irradiapy.io.xyzreader import XYZReader
from irradiapy.io.xyzwriter import XYZWriter


class DefectsFinderOld:
    """Class to find and analyze defects in crystal structures (deprecated).

    Warning
    -------
    This class is was used to analyse the defects on CascadesDB, but it is not
    the best option. The class `DefectsFinder` is the recommended one as it is
    optimized. This class is no longer maintained.

    Note
    ----
    Even if an atom is very far away from the simulation box, the closest lattice
    site will be asigned to it. This is a limitation of the current implementation.
    A cutoff distance should be implemented to avoid this and consider the atom as
    free.

    Attributes
    ----------
    nx, ny, nz : int
        Dimensions of the lattice.
    perx, pery, perz : bool
        Periodic boundary conditions along x, y, and z axes.
    dtype : npt.DTypeLike or None
        Data type for atomic positions.
    norm_cell : DefectsFinder._Cell
        Normalized cell for affine transformations.
    rng : np.random.Generator
        Random number generator.
    """

    # Full atom, used only by DefectsFinder
    fatom = np.dtype(
        [("type", int), ("pos", float, 3), ("index", int, 4), ("modular", float, 3)]
    )

    # Unit cell, used only by DefectsFinder
    ucell = np.dtype([("pos", float, 3), ("index", int, 4)])

    def __init__(
        self,
        nx: int,
        ny: int,
        nz: int,
        perx: bool,
        pery: bool,
        perz: bool,
        dtype: Optional[npt.DTypeLike] = None,
        seed: Optional[int] = None,
    ) -> None:
        """Initializes the `DefectsFinder` object.

        Parameters
        ----------
        nx, ny, nz : int
            Dimensions of the lattice.
        perx, pery, perz : bool
            Periodic boundary conditions along x, y, and z axes.
        dtype : npt.DTypeLike or None, optional
            Data type for atomic positions.
        seed : int or None, optional
            Seed for the random number generator.
        """
        self.nx, self.ny, self.nz = nx, ny, nz
        self.perx, self.pery, self.perz = perx, pery, perz
        self.dtype = dtype
        self.norm_cell = self.__get_norm_cell()
        self.rng = np.random.default_rng(seed)

    def __get_norm_cell(self) -> "DefectsFinder._Cell":  # type: ignore
        """Calculates the normalized cell for affine transformations.

        Returns
        -------
        DefectsFinder._Cell
            Normalized cell with lattice parameters equal to 1.
        """
        vectors = np.diag([self.nx - 0.5, self.ny - 0.5, self.nz - 0.5])
        origin = np.zeros(3)
        return self._Cell(origin, vectors)

    def __get_curr_cell(
        self, atoms: "DefectsFinder.fatom"  # type: ignore
    ) -> "DefectsFinder._Cell":  # type: ignore
        """Calculates the current cell based on atomic positions.

        Parameters
        ----------
        atoms : dtypes.atom
            Atomic positions.

        Returns
        -------
        DefectsFinder._Cell
            Current cell based on atomic positions.
        """
        origin = np.min(atoms["pos"], axis=0)
        vectors = np.diag(np.max(atoms["pos"], axis=0) - origin)
        return self._Cell(origin, vectors)

    def __affine_transformation(
        self,
        curr_cell: "DefectsFinder._Cell",  # type: ignore
        tar_cell: "DefectsFinder._Cell",  # type: ignore
        atoms: Union["DefectsFinder.fatom", np.ndarray],  # type: ignore
    ) -> Union["DefectsFinder.fatom", np.ndarray]:  # type: ignore
        """Applies affine transformation to atomic positions.

        Parameters
        ----------
        curr_cell : DefectsFinder._Cell
            Current cell.
        tar_cell : DefectsFinder._Cell
            Target cell.
        atoms : dtypes.atom
            Atomic positions.

        Returns
        -------
        dtypes.atom
            Transformed atomic positions.
        """
        if len(atoms) == 0:
            return atoms
        atoms["pos"] -= curr_cell.origin
        transform = np.dot(tar_cell.vectors, np.linalg.inv(curr_cell.vectors))
        atoms["pos"] = np.dot(atoms["pos"], transform)
        atoms["pos"] += tar_cell.origin
        return atoms

    def rescale_translate_rotate(
        self,
        atoms: Union["DefectsFinder.fatom", np.ndarray],  # type: ignore
        pka_pos: np.ndarray,
        pka_theta: float,
        pka_phi: float,
        a0: Optional[Union[float, np.ndarray]] = None,
        a1: Optional[Union[float, np.ndarray]] = None,
    ) -> Union["DefectsFinder.fatom", np.ndarray]:  # type: ignore
        """Rescales, translates, and rotates the positions of atoms.

        Parameters
        ----------
        atoms : dict
            Dictionary containing atomic positions under the key "pos".
        pka_pos : array-like
            Position vector of the primary knock-on atom (PKA).
        pka_theta : float
            Polar angle (in radians) for the PKA direction.
        pka_phi : float
            Azimuthal angle (in radians) for the PKA direction.
        a0 : float or array-like
            Initial scaling factor or vector.
        a1 : float or array-like
            Final scaling factor or vector.

        Returns
        -------
        dict
            Updated dictionary with rescaled, translated, and rotated atomic positions.
        """
        # Translate
        atoms["pos"] -= pka_pos
        # Rotation matrix, align with PKA initial direction
        xaxis = np.array([1.0, 0.0, 0.0])
        pka_dir = np.array(
            [
                np.sin(pka_theta) * np.cos(pka_phi),
                np.sin(pka_theta) * np.sin(pka_phi),
                np.cos(pka_theta),
            ]
        )
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            transform = R.align_vectors([xaxis], [pka_dir])[0].as_matrix()
        # Scaling matrix
        scaling = a1 / a0
        scaling_matrix = (
            np.diag([scaling] * 3) if isinstance(scaling, float) else np.diag(scaling)
        )
        # Scale-rotate matrix
        transform = transform @ scaling_matrix
        # Apply scaling and rotation
        atoms["pos"] = atoms["pos"] @ transform.T
        return atoms

    def __generate_assignments(self) -> dict:
        """Generates the corresponding lattice in index coordinates (bcc lattice).

        Returns
        -------
        dict
            Dictionary with lattice positions and assigned atoms.

        Examples
        --------
        assignments = {
            (1, 1, 1, 1): [123, 23],
            ...
        }
        lattice position with indices (1, 1, 1, 1) has two atoms assigned, whose
        numbers are 123 and 23
        """
        assignments = {
            (xi, yi, zi, sub): []
            for xi in range(self.nx)
            for yi in range(self.ny)
            for zi in range(self.nz)
            for sub in (0, 1)
        }
        return assignments

    def find(
        self,
        ifile_path: Path,
        ofile_path: Path,
        pka_pos: Optional[np.ndarray] = None,
        pka_theta: Optional[float] = None,
        pka_phi: Optional[float] = None,
        a0: Optional[float] = None,
        a1: Optional[float] = None,
    ) -> np.ndarray:  # type: ignore
        """Finds defects in `ifile_path` and saves the info in `ofile_path`.

        Parameters
        ----------
        ifile_path : Path
            Input file path.
        ofile_path : Path
            Output file path.
        pka_pos : None or npt.NDArray[np.float64], optional
            Position vector of the primary knock-on atom (PKA).
        pka_theta : None or float, optional
            Polar angle (in radians) for the PKA direction.
        pka_phi : None or float, optional
            Azimuthal angle (in radians) for the PKA direction.
        a0 : None or float, optional
            Initial scaling factor.
        a1 : None or float, optional
            Final scaling factor.

        Returns
        -------
        dtypes.defect
            Array of defects.
        """
        reader = XYZReader(ifile_path, self.dtype)
        atoms0 = list(reader)[0]
        natoms = len(atoms0)
        atoms = np.zeros(natoms, dtype=self.fatom)
        atoms["type"] = atoms0["type"]
        atoms["pos"] = atoms0["pos"]
        del atoms0
        print(f"Number of atoms: {natoms}")
        # Affine transformation
        curr_cell = self.__get_curr_cell(atoms)
        atoms = self.__affine_transformation(curr_cell, self.norm_cell, atoms)
        # Unit cell coordinates (bcc lattice)
        ucell = np.array(
            [
                ([0.0, 0.0, 0.0], [0, 0, 0, 0]),
                ([1.0, 0.0, 0.0], [1, 0, 0, 0]),
                ([1.0, 1.0, 0.0], [1, 1, 0, 0]),
                ([0.0, 1.0, 0.0], [0, 1, 0, 0]),
                ([0.0, 0.0, 1.0], [0, 0, 1, 0]),
                ([1.0, 0.0, 1.0], [1, 0, 1, 0]),
                ([1.0, 1.0, 1.0], [1, 1, 1, 0]),
                ([0.0, 1.0, 1.0], [0, 1, 1, 0]),
                ([0.5, 0.5, 0.5], [0, 0, 0, 1]),
            ],
            dtype=self.ucell,
        )
        # Atom index and modular coordinates
        index, modular = np.divmod(atoms["pos"], 1.0)
        atoms["index"][:, :-1], atoms["modular"] = index, modular
        del index, modular
        # Array of distances:
        # each row (atom) contains the distance to each unit cell position
        dist = np.sum(np.square(atoms["modular"][:, np.newaxis] - ucell["pos"]), axis=2)
        # Get assigned lattice index position for the atom
        atoms["index"] += ucell["index"][np.argmin(dist, axis=1)]
        # Periodic boundary conditions
        if self.perx:
            atoms["index"][:, 0][atoms["index"][:, 0] == self.nx] = 0
            atoms["index"][:, 0][atoms["index"][:, 0] == -1] = self.nx - 1
        if self.pery:
            atoms["index"][:, 1][atoms["index"][:, 1] == self.ny] = 0
            atoms["index"][:, 1][atoms["index"][:, 1] == -1] = self.ny - 1
        if self.perz:
            atoms["index"][:, 2][atoms["index"][:, 2] == self.nz] = 0
            atoms["index"][:, 2][atoms["index"][:, 2] == -1] = self.nz - 1
        # Create all lattice positions and assign a lattice position to each atom
        assignments = self.__generate_assignments()
        print(f"Lattice sites: {len(assignments)}")
        for i in range(natoms):
            assignments[tuple(atoms["index"][i])].append(i)
        # Vacancies and interstitials identification
        defects = np.empty(0, dtypes.defect)
        for key, value in assignments.items():
            # lattice site cartesian coordinates
            site_pos = np.array(key[:-1]) + key[-1] * 0.5
            # no atoms assigned > vacancy
            if len(value) == 0:
                defects = np.concatenate(
                    (defects, np.array([(0, site_pos)], dtype=dtypes.defect))
                )
            elif len(value) > 1:
                # do not consider as defect the closest atom to the lattice site
                dist = np.sum(np.square(site_pos - atoms[value]["pos"]))
                mask = np.full(len(value), True)
                mask[np.argmin(dist)] = False
                inters = np.empty(len(value) - 1, dtypes.defect)
                inters["type"] = atoms[value]["type"][mask]
                inters["pos"] = atoms[value]["pos"][mask]
                defects = np.concatenate((defects, inters))
        # Undo affine transformation
        defects = self.__affine_transformation(self.norm_cell, curr_cell, defects)
        if pka_pos is None or pka_theta is None or pka_phi is None:
            print("PKA parameters or lattice parameter are ungiven.")
            defects["pos"] -= np.mean(defects["pos"], axis=0)
            if a0 is not None and a1 is not None:
                defects["pos"] *= a1 / a0
        else:
            defects = self.rescale_translate_rotate(
                defects, pka_pos, pka_theta, pka_phi, a0, a1
            )
        nvacs = len(np.where(defects["type"] == 0)[0])
        nsias = len(defects) - nvacs
        print(f"Number of interstitials: {nsias}")
        print(f"Number of vacancies: {nvacs}")
        writer = XYZWriter(ofile_path)
        writer.save(defects)
        writer.close()
        return defects

    @dataclass
    class _Cell:
        """Helper class that contains data of the simulation cell.

        Attributes
        ----------
        origin : npt.NDArray[np.float64]
            Origin of the cell.
        vectors : npt.NDArray[np.float64]
            Vectors defining the cell.
        """

        origin: np.ndarray
        vectors: np.ndarray



"""This module contains the `DamageDB` class."""

import warnings
from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from scipy.spatial.transform import Rotation
from sklearn.decomposition import PCA

from irradiapy import dpa, dtypes, materials
from irradiapy.io.xyzreader import XYZReader


@dataclass
class DamageDB:
    """Class used to reconstruct the damage produced by a PKA from a database of MD debris.

    Attributes
    ----------
    dir_mddb : Path
        Directory of the MD debris database.
    mat_pka : materials.Material
        PKA material.
    mat_target : materials.Material
        Target material.
    compute_tdam : bool
        Whether to apply Lindhard's formula to the recoil energy. It should be `True` for
        MD simulations without electronic stopping.
    dpa_mode : dpa.DpaMode
        Mode for dpa calculation.
    force_lss : bool, optional
        If True, force the use of the Lindhard formula for damage energy calculation. Default is
        False.
    seed : int, optional
        Random seed for random number generator. Default is 0.
    """

    dir_mddb: Path
    compute_tdam: bool
    mat_pka: materials.Material
    mat_target: materials.Material
    dpa_mode: dpa.DpaMode
    force_lss: bool = False
    seed: int = 0
    __rng: np.random.Generator = field(init=False)
    __calc_nd: callable = field(init=False)
    __files: dict[float, list[Path]] = field(init=False)
    __energies: np.ndarray = field(init=False)
    __nenergies: int = field(init=False)

    def __post_init__(self) -> None:
        self.__rng = np.random.default_rng(self.seed)
        # Scan the database
        self.__files = {
            float(folder.name): list(folder.iterdir())
            for folder in self.dir_mddb.iterdir()
            if folder.is_dir()
        }
        self.__energies = np.array(sorted(self.__files.keys(), reverse=True))
        self.__nenergies = len(self.__energies)
        # PKA energy to damage energy conversion
        self.__compute_damage_energy = lambda x: dpa.compute_damage_energy(
            x, self.mat_pka, self.mat_target, force_lss=self.force_lss
        )
        # Select the dpa model for residual energy
        if self.dpa_mode == dpa.DpaMode.NRT:
            self.__calc_nd = lambda x: np.round(
                dpa.calc_nrt_dpa(x, self.mat_target)
            ).astype(np.int32)
        elif self.dpa_mode == dpa.DpaMode.ARC:
            self.__calc_nd = lambda x: np.round(
                dpa.calc_arc_dpa(x, self.mat_target)
            ).astype(np.int32)
        elif self.dpa_mode == dpa.DpaMode.FERARC:
            self.__calc_nd = lambda x: np.round(
                dpa.calc_fer_arc_dpa(x, self.mat_target)
            ).astype(np.int32)
        else:
            raise ValueError("Invalid dpa mode")

    def __get_files(self, pka_e: float) -> tuple[dict[float, list[Path]], int]:
        """Get cascade files and number of residual FP for a given PKA energy.

        Parameters
        ----------
        pka_e : float
            PKA energy.

        Returns
        -------
        tuple[dict[float, list[Path]], int]
            Dictionary of selected paths and number of residual FP.
        """
        # Decompose the PKA energy into cascades and residual energy
        residual_energy = (
            self.__compute_damage_energy(pka_e) if self.compute_tdam else pka_e
        )
        cascade_counts = np.zeros(self.__nenergies, dtype=int)
        for i, energy in enumerate(self.__energies):
            cascade_counts[i], residual_energy = divmod(residual_energy, energy)
        # Select the files for each energy
        if residual_energy > 0:
            residual_energy = self.__compute_damage_energy(residual_energy)
        debris_files = {
            energy: self.__rng.choice(self.__files[energy], cascade_counts[i])
            for i, energy in enumerate(self.__energies)
        }
        # Get the number of residual FP
        nfp = self.__calc_nd(residual_energy)
        return debris_files, nfp

    def get_pka_debris(
        self, pka_e: float, pka_pos: np.ndarray, pka_dir: np.ndarray
    ) -> np.ndarray:
        """Get PKA debris from its energy position, and direction.

        Parameters
        ----------
        pka_e : float
            PKA energy.
        pka_pos : np.ndarray
            PKA position.
        pka_dir : np.ndarray
            PKA direction.

        Returns
        -------

        np.ndarray
            Defects after the cascades.
        """
        files, nfp = self.__get_files(pka_e)
        # Get the maximum energy available in the database for the given PKA.
        # If no energy is available, return zero to place only FP.
        db_emax = next(
            (energy for energy in self.__energies if len(files[energy])), 0.0
        )
        # Possible to get cascades from the database
        if db_emax > 0.0:
            defects = self.__process_highest_energy_cascade(
                files, db_emax, pka_pos, pka_dir
            )
            parallelepiped = self.__get_parallelepiped(defects)
            defects = self.__place_other_debris(files, defects, parallelepiped)
            if nfp:
                defects = self.__place_fps_in_parallelepiped(
                    defects, nfp, parallelepiped
                )
            return defects
        # If no energy is available, generate FP only
        if nfp:
            return self.__place_fps_in_sphere(nfp, pka_pos, pka_dir)
        return np.empty(0, dtype=dtypes.defect)

    def __process_highest_energy_cascade(
        self,
        files: dict,
        db_emax: float,
        pka_pos: np.ndarray,
        pka_dir: np.ndarray,
    ) -> np.ndarray:
        """Process the highest energy cascade.

        Parameters
        ----------
        files : dict
            Dictionary of files for each energy.
        db_emax : float
            Energy of the highest energy cascade.
        pka_pos : np.ndarray
            PKA position.
        pka_dir : np.ndarray
            PKA direction.

        Returns
        -------
        np.ndarray
            Defects after the highest energy cascade.
        """
        file = files[db_emax][0]
        files[db_emax] = np.delete(files[db_emax], 0)
        defects = next(iter(XYZReader(file)))
        xaxis = np.array([1.0, 0.0, 0.0])
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            transform = Rotation.align_vectors([pka_dir], [xaxis])[0].as_matrix()
        defects["pos"] = defects["pos"] @ transform.T
        defects["pos"] += pka_pos
        return defects

    def __place_other_debris(
        self,
        files: dict,
        defects: np.ndarray,
        parallelepiped: tuple[PCA, np.ndarray, np.ndarray],
    ) -> np.ndarray:
        """Place other debris in the parallelepiped.

        Parameters
        ----------
        files : dict
            Dictionary of files for each energy.
        defects : np.ndarray
            Defects after the highest energy cascade.
        parallelepiped : tuple[PCA, np.ndarray, np.ndarray]
            Parallelepiped definition.

        Returns
        -------
        np.ndarray
            Defects after placing the other debris.
        """
        for energy in self.__energies:
            for file0 in files[energy]:
                defects0 = next(iter(XYZReader(file0)))
                defects0["pos"] -= np.mean(defects0["pos"], axis=0)
                defects0["pos"] = Rotation.random(rng=self.__rng).apply(defects0["pos"])
                pos0 = self.__get_parallelepiped_points(*parallelepiped, 1)
                defects0["pos"] += pos0
                defects = np.concatenate((defects, defects0))
        return defects

    def __place_fps_in_parallelepiped(
        self,
        defects: np.ndarray,
        nfp: int,
        parallelepiped: tuple[PCA, np.ndarray, np.ndarray],
    ) -> np.ndarray:
        """Place FPs anywhere in the parallelepiped.

        Parameters
        ----------
        defects : np.ndarray
            Defects after placing the other debris.
        nfp : int
            Number of FPs.
        parallelepiped : tuple[PCA, np.ndarray, np.ndarray]
            Parallelepiped definition.

        Returns
        -------
        np.ndarray
            Defects after placing the FPs.
        """
        defects0 = np.zeros(2 * nfp, dtype=dtypes.defect)
        defects0["type"][:nfp] = self.mat_target.atomic_number
        defects0["type"][nfp:] = 0
        defects0["pos"][:nfp, 0] = self.mat_target.dist_fp / 2.0
        defects0["pos"][:nfp] = Rotation.random(nfp, self.__rng).apply(
            defects0["pos"][:nfp]
        )
        defects0["pos"][nfp:] = -defects0["pos"][:nfp]
        pos0 = self.__get_parallelepiped_points(*parallelepiped, nfp)
        defects0["pos"][:nfp] += pos0
        defects0["pos"][nfp:] += pos0
        return np.concatenate((defects, defects0))

    def __place_fps_in_sphere(
        self, nfp: int, pka_pos: np.ndarray, pka_dir: np.ndarray
    ) -> np.ndarray:
        """Generate FPs in a sphere.

        Parameters
        ----------
        nfp : int
            Number of FPs.
        pka_pos : np.ndarray
            PKA position.
        pka_dir : np.ndarray
            PKA direction.

        Returns
        -------
        np.ndarray
            Defects after generating.
        """
        defects = np.zeros(2 * nfp, dtype=dtypes.defect)
        defects["type"][:nfp] = self.mat_target.atomic_number
        defects["type"][nfp:] = 0
        defects["pos"][:nfp, 0] = self.mat_target.dist_fp / 2.0
        defects["pos"][:nfp] = Rotation.random(nfp, rng=self.__rng).apply(
            defects["pos"][:nfp]
        )
        defects["pos"][nfp:] = -defects["pos"][:nfp]

        random = self.__rng.random((nfp, 3))
        theta = np.arccos(2.0 * random[:, 0] - 1.0)
        phi = 2.0 * np.pi * random[:, 1]
        radius = nfp * self.mat_target.dist_fp / 2.0
        r = radius * np.cbrt(random[:, 2])
        points = np.empty((nfp, 3))
        points[:, 0] = r * np.sin(theta) * np.cos(phi)
        points[:, 1] = r * np.sin(theta) * np.sin(phi)
        points[:, 2] = r * np.cos(theta)

        defects["pos"][:nfp] += points
        defects["pos"][nfp:] += points
        defects["pos"] += pka_pos + pka_dir * radius
        return defects

    def __get_parallelepiped(self, atoms: np.ndarray) -> tuple:
        """
        Define a parallelepiped from the atomic positions using PCA.

        Parameters
        ----------
        atoms : np.ndarray
            Atomic positions.

        Returns
        -------
        tuple
            PCA object, minimum PCA coordinates, maximum PCA coordinates.
        """
        pca = PCA(n_components=3)
        pca.fit(atoms["pos"])
        atoms_pca = pca.transform(atoms["pos"])
        min_pca = np.min(atoms_pca, axis=0)
        max_pca = np.max(atoms_pca, axis=0)
        return pca, min_pca, max_pca

    def __get_parallelepiped_points(
        self,
        pca: PCA,
        min_pca: np.ndarray,
        max_pca: np.ndarray,
        npoints: int,
    ) -> np.ndarray:
        """
        Generate random points within a parallelepiped.

        Parameters
        ----------
        pca : PCA
            PCA object.
        min_pca : np.ndarray
            Minimum PCA coordinates.
        max_pca : np.ndarray
            Maximum PCA coordinates.
        npoints : int
            Number of points to generate.

        Returns
        -------
        np.ndarray
            Random points within the parallelepiped.
        """
        random_points_pca = self.__rng.uniform(min_pca, max_pca, size=(npoints, 3))
        return pca.inverse_transform(random_points_pca)



"""This module contains math utilities for the irradiapy package."""

# pylint: disable=unbalanced-tuple-unpacking

from typing import Callable, Optional, Union

import numpy as np
from scipy.optimize import curve_fit


def lorentzian(
    xs: np.ndarray,
    x_peak: float,
    linewidth: float,
    amplitude: float,
    asymmetry: float,
) -> Union[float, np.ndarray]:
    """Evaluate a Lorentzian function.

    Parameters
    ----------
    xs : np.ndarray
        Where to evaluate the function.
    x_peak : float
        Position with maximum value.
    linewidth : float
        Linewidth.
    amplitude : float
        Maximum amplitude.
    asymmetry : float
        Asymmetry.

    Returns
    -------
    Union[float, np.ndarray]
        Evaluated Lorentzian function.

    References
    ----------
    See https://doi.org/10.1016/j.nimb.2021.05.014
    """
    delta_x = xs - x_peak
    linewidth_sq = linewidth**2
    exp_term = np.exp(asymmetry * delta_x)
    alpha = (1.0 + exp_term) ** 2
    alpha_quarter = alpha / 4.0
    exponent = -alpha_quarter * delta_x**2 / (2.0 * linewidth_sq)
    return amplitude * alpha_quarter * np.exp(exponent)


def fit_lorentzian(
    xs: np.ndarray,
    ys: np.ndarray,
    p0: Optional[np.ndarray] = None,
    asymmetry: float = 1.0,
) -> tuple[np.ndarray, np.ndarray, Callable[[np.ndarray], np.ndarray]]:
    """Fit data to a Lorentzian function.

    Parameters
    ----------
    xs : np.ndarray
        X values where the function is evaluated.
    ys : np.ndarray
        Y values at the given xs.
    p0 : np.ndarray, optional
        Initial guess of fit parameters. If None, a guess is generated. Default is None.
    asymmetry : float, optional
        Bound for the asymmetry fit parameter. Fit will be done in (-asymmetry, asymmetry). Default is 1.0.

    Returns
    -------
    popt : np.ndarray
        Optimal values for the parameters.
    pcov : np.ndarray
        Covariance of popt.
    fit_function : Callable[[np.ndarray], np.ndarray]
        Function that evaluates the fitted Lorentzian.
    """
    if p0 is None:
        peak_index = np.argmax(ys)
        x_start = xs[0]
        x_end = xs[-1]
        sigma_guess = 0.5 * (x_end - x_start)
        p0 = np.array(
            [
                xs[peak_index],
                sigma_guess,
                ys[peak_index],
                0.0,
            ]
        )
    x_start = xs[0]
    x_end = xs[-1]
    x_sum = x_start + x_end
    popt, pcov = curve_fit(
        lorentzian,
        xs,
        ys,
        p0=p0,
        bounds=(
            [x_start, 0.0, ys.min(), -asymmetry],
            [x_end, x_sum, ys.max(), asymmetry],
        ),
    )

    def fit_function(xs_fit: np.ndarray) -> np.ndarray:
        return lorentzian(xs_fit, *popt)

    return popt, pcov, fit_function


def gaussian(
    xs: np.ndarray,
    x_peak: float,
    linewidth: float,
    amplitude: float,
    asymmetry: float,
) -> Union[float, np.ndarray]:
    """Evaluate a Gaussian function.

    Parameters
    ----------
    xs : np.ndarray
        Where to evaluate the function.
    x_peak : float
        Position with maximum value.
    linewidth : float
        Linewidth.
    amplitude : float
        Maximum amplitude.
    asymmetry : float
        Asymmetry.

    Returns
    -------
    Union[float, np.ndarray]
        Evaluated Gaussian function.

    References
    ----------
    See https://doi.org/10.1016/j.nimb.2021.05.014
    """
    delta_x = xs - x_peak
    linewidth_sq = linewidth**2
    exp_term = np.exp(asymmetry * delta_x)
    alpha = (1.0 + exp_term) ** 2
    alpha_quarter = alpha / 4.0
    exponent = -alpha_quarter * delta_x**2 / (2.0 * linewidth_sq)
    return amplitude * alpha_quarter * np.exp(exponent)


def fit_gaussian(
    xs: np.ndarray,
    ys: np.ndarray,
    p0: Optional[np.ndarray] = None,
    asymmetry: float = 1.0,
) -> tuple[np.ndarray, np.ndarray, Callable[[np.ndarray], np.ndarray]]:
    """Fit data to a Gaussian function.

    Parameters
    ----------
    xs : np.ndarray
        X values where the function is evaluated.
    ys : np.ndarray
        Y values at the given xs.
    p0 : np.ndarray, optional
        Initial guess of fit parameters. If None, a guess is generated. Default is None.
    asymmetry : float, optional
        Bound for the asymmetry fit parameter. Fit will be done in (-asymmetry, asymmetry). Default is 1.0.

    Returns
    -------
    popt : np.ndarray
        Optimal values for the parameters.
    pcov : np.ndarray
        Covariance of popt.
    fit_function : Callable[[np.ndarray], np.ndarray]
        Function that evaluates the fitted Gaussian.
    """
    if p0 is None:
        peak_index = np.argmax(ys)
        x_start = xs[0]
        x_end = xs[-1]
        sigma_guess = 0.5 * (x_end - x_start)
        p0 = np.array(
            [
                xs[peak_index],
                sigma_guess,
                ys[peak_index],
                0.0,
            ]
        )
    x_start = xs[0]
    x_end = xs[-1]
    x_sum = x_start + x_end
    popt, pcov = curve_fit(
        gaussian,
        xs,
        ys,
        p0=p0,
        bounds=(
            [x_start, 0.0, ys.min(), -asymmetry],
            [x_end, x_sum, ys.max(), asymmetry],
        ),
    )

    def fit_function(xs_fit: np.ndarray) -> np.ndarray:
        return gaussian(xs_fit, *popt)

    return popt, pcov, fit_function



"""This module contains the `Trimdat` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Trimdat(SRIMFile):
    """Class to handle `TRIM.DAT` file."""

    def process_file(self, trimdat_path: Path) -> None:
        """Processes `TRIM.DAT` file.

        Parameters
        ----------
        trimdat_path : Path
            `TRIM.DAT` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE trimdat"
                "(ion_numb INTEGER, atom_numb INTEGER, energy REAL, depth REAL,"
                "y REAL, z REAL, cosx REAL, cosy REAL, cosz REAL)"
            )
        )
        with open(trimdat_path, "r", encoding="utf-8") as file:
            for _ in range(10):
                file.readline()
            cur = self.cursor()
            for line in file:
                data = line[:-1].split()
                ion_numb = int(data[0])
                atom_numb = int(data[1])
                energy = float(data[2])
                depth = float(data[3])
                y = float(data[4])
                z = float(data[5])
                cosx = float(data[6])
                cosy = float(data[7])
                cosz = float(data[8])
                cur.execute(
                    (
                        "INSERT INTO trimdat"
                        "(ion_numb, atom_numb, energy, depth, y, z, cosx, cosy, cosz)"
                        "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?)"
                    ),
                    [ion_numb, atom_numb, energy, depth, y, z, cosx, cosy, cosz],
                )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads trimdat data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM trimdat {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the trimdat table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge with.
        """
        nions = self.srim.nions
        cur = self.cursor()
        cur.execute(f"ATTACH DATABASE '{srimdb2.db_path}' AS srimdb2")
        cur.execute(
            (
                "INSERT INTO trimdat"
                "(ion_numb, atom_numb, energy, depth, y, z, cosx, cosy, cosz)"
                "SELECT ion_numb + ?, atom_numb, energy, depth, y, z, cosx, cosy, cosz "
                "FROM srimdb2.trimdat"
            ),
            (nions,),
        )
        self.srim.commit()
        cur.execute("DETACH DATABASE srimdb2")
        cur.close()



"""irradiapy package"""

import matplotlib.pyplot as plt

from irradiapy import dpa, dtypes, io, materials, srimpy, analysis
from irradiapy.damagedb import DamageDB


def use_style(latex: bool = False) -> None:
    """Set the style for matplotlib plots.

    It uses the colour universal design (CUD) palette for colour-blind friendly plots.

    Parameters
    ----------
    latex : bool, optional
        If True, use LaTeX for text rendering in plots (slower). Default is False. I might require
        other software to be installed on your system.
    """
    if latex:
        plt.style.use("irradiapy.styles.latex")
    else:
        plt.style.use("irradiapy.styles.nolatex")
        plt.style.use("irradiapy.styles.nolatex")



"""Utility functions related to PKA for SRIM data."""

from itertools import combinations
from pathlib import Path
from typing import Callable, Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.image import NonUniformImage
from scipy.optimize import curve_fit

from irradiapy.srimpy.srimdb import SRIMDB


def plot_pka_distribution(
    srimdb: SRIMDB,
    bins: int = 100,
    plot_path: Optional[Path] = None,
    dpi: int = 300,
    fit_path: Optional[Path] = None,
) -> Callable:
    """Plot the PKA energy distribution and tries to fit it.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    bins : int, optional
        Energy bins, by default 100. The fit will be done over non-empty bins.
    plot_path : Path, optional
        Output path for the plot, by default None. If None, the plot is shown.
    dpi : int, optional
        Dots per inch for the plot, by default 300.
    fit_path : Path, optional
        Output path for the fit parameters, by default None.

    Returns
    -------
    Callable
        Scaling law function.
    """
    # Read
    nions = srimdb.nions
    pka_es = np.array(list(srimdb.collision.read(what="recoil_energy")))
    # Histogram
    pka_hist, pka_edges = np.histogram(pka_es, bins=bins)
    pka_hist = pka_hist / nions
    pka_centers = pka_edges[:-1] + (pka_edges[1:] - pka_edges[:-1]) / 2.0
    # Fit
    try:
        # pylint: disable=unbalanced-tuple-unpacking
        popt, _ = curve_fit(
            lambda x, a, b: a + b * x,
            np.log10(pka_centers[pka_hist > 0] / 1e3),
            np.log10(pka_hist[pka_hist > 0]),
        )
        a, s = 10.0 ** popt[0], -popt[1]

        def curve(x):
            return a / x**s

        fit = True
    except Exception as exc:  # pylint: disable=broad-except
        print(f"Fit failed: {exc}")
    if fit and fit_path:
        with open(fit_path, "w", encoding="utf-8") as file:
            file.write(f"PKA energy scaling law (x in eV)\nA/x**S\nA S\n{a}, {s}\n")
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec()
    # Scatter
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"$E_{PKA}$ (keV)")
    ax.set_ylabel("Counts per ion")
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.scatter(pka_centers / 1e3, pka_hist)
    if fit:
        ax.plot(
            pka_centers / 1e3,
            curve(pka_centers / 1e3),
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
    # Finish
    fig.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()
    return curve


def plot_energy_depth(
    srimdb: SRIMDB,
    depth_bins: int = 100,
    pka_ebins: int = 100,
    pka_e_max: float = 200,
    plot_high_path: Optional[Path] = None,
    plot_low_path: Optional[Path] = None,
    dpi: int = 300,
) -> None:
    """Plots the depth-energy distribution of PKAs.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    depth_bins : int, optional
        Number of bins for depth histogram, by default 100.
    pka_ebins : int, optional
        Number of bins for PKA energy histogram, by default 100.
    pka_e_max : float, optional
        Maximum PKA energy, by default 200.
    plot_high_path : Path, optional
        Output path for the high energy plot, by default None. If None, the plot is shown.
    plot_low_path : Path, optional
        Output path for the low energy plot, by default None. If None, the plot is shown.
    dpi : int, optional
        Dots per inch for the plot, by default 300.
    """
    # Read
    nions = srimdb.nions
    data = np.array(list(srimdb.collision.read(what="depth, recoil_energy")))
    depths, pka_es = data[:, 0], data[:, 1]
    # Low energy, linear
    depth_edges = np.histogram_bin_edges(depths, bins=depth_bins)
    pka_e_edges = np.histogram_bin_edges(
        pka_es, bins=pka_ebins, range=(pka_es.min(), pka_e_max)
    )
    hist, _, _ = np.histogram2d(depths, pka_es, bins=[depth_edges, pka_e_edges])
    hist /= nions
    depth_centers = depth_edges[:-1] + (depth_edges[1:] - depth_edges[:-1]) / 2.0
    pka_e_centers = pka_e_edges[:-1] + (pka_e_edges[1:] - pka_e_edges[:-1]) / 2.0
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[0.95, 0.05])
    cmap = plt.cm.get_cmap("viridis")
    cmap.set_under(plt.rcParams["axes.facecolor"])
    # Color map
    ax = fig.add_subplot(gs[0, 0])
    ax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_xlabel(r"$E_{PKA}$ (eV) (low energies)")
    ax.set_xlim(pka_e_edges[[0, -1]])
    ax.set_ylim(depth_edges[[0, -1]])
    im = NonUniformImage(
        ax, cmap=cmap, extent=(*pka_e_edges[[0, -1]], *depth_edges[[0, -1]])
    )
    im.set_clim(vmin=1 / nions)
    im.set_data(pka_e_centers, depth_centers, hist)
    ax.add_image(im)
    # Color bar
    cax = fig.add_subplot(gs[0, 1])
    fig.colorbar(im, cax, label="Counts per ion")
    plt.tight_layout()
    if plot_low_path:
        plt.savefig(plot_low_path, dpi=dpi)
    else:
        plt.show()
    plt.close()
    # All energies, log
    depth_edges = np.histogram_bin_edges(depths, bins=depth_bins)
    pka_e_edges = np.histogram_bin_edges(pka_es, bins=pka_ebins)
    hist, _, _ = np.histogram2d(depths, pka_es, bins=[depth_edges, pka_e_edges])
    hist /= nions
    depth_centers = depth_edges[:-1] + (depth_edges[1:] - depth_edges[:-1]) / 2.0
    pka_e_centers = pka_e_edges[:-1] + (pka_e_edges[1:] - pka_e_edges[:-1]) / 2.0
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[0.95, 0.05])
    # Color map
    ax = fig.add_subplot(gs[0, 0])
    ax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_xlabel(r"$E_{PKA}$ (keV)")
    ax.set_xlim(pka_e_edges[[0, -1]] / 1e3)
    ax.set_ylim(depth_edges[[0, -1]])
    im = NonUniformImage(
        ax,
        cmap=cmap,
        norm="log",
        extent=(*pka_e_edges[[0, -1]] / 1e3, *depth_edges[[0, -1]]),
    )
    im.set_clim(vmin=1 / nions)
    im.set_data(pka_e_centers / 1e3, depth_centers, hist)
    ax.add_image(im)
    # Color bar
    cax = fig.add_subplot(gs[0, 1])
    fig.colorbar(im, cax, label="Counts per ion")
    # Finish
    plt.tight_layout()
    if plot_high_path:
        plt.savefig(plot_high_path, dpi=dpi)
    else:
        plt.show()
    plt.close()


def plot_distances(
    srimdb: SRIMDB,
    pka_e_lim: float = 5e3,
    dist_bins: int = 100,
    energy_bins: int = 100,
    plot_path: Optional[Path] = None,
    dpi: int = 300,
) -> None:
    """Plots a 2D histogram of pairwise distances and sum of PKA energies for each ion.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    pka_elim : float, optional
        Minimum recoil energy to consider, by default 5e3.
    dist_bins : int, optional
        Number of bins for the distance histogram, by default 100.
    energy_bins : int, optional
        Number of bins for the energy histogram, by default 100.
    plot_path : Path, optional
        Output path for the plot, by default None. If None, the plot is shown.
    dpi : int, optional
        Dots per inch for the plot, by default 300.
    """
    nions = srimdb.nions
    distances = []
    energies = []
    # Get pairwise distances and energies for each ion
    for nion in range(1, nions + 1):
        data = np.array(
            list(
                srimdb.collision.read(
                    what="depth, y, z, recoil_energy",
                    condition=f"WHERE ion_numb = {nion} AND recoil_energy >= {pka_e_lim}",
                )
            )
        )
        if len(data):
            pos = data[:, :3]
            pka_e = data[:, 3]
            for i, j in combinations(range(len(pos)), 2):
                distance = np.linalg.norm(pos[i] - pos[j])
                energy = pka_e[i] + pka_e[j]
                distances.append(distance)
                energies.append(energy)
    distances = np.array(distances)
    energies = np.array(energies) / 1e3
    # Histogram
    dist_edges = np.histogram_bin_edges(distances, bins=dist_bins)
    energies_edges = np.histogram_bin_edges(energies, bins=energy_bins)
    hist, _, _ = np.histogram2d(distances, energies, bins=[dist_edges, energies_edges])
    hist /= nions
    dist_centers = dist_edges[:-1] + (dist_edges[1:] - dist_edges[:-1]) / 2.0
    energies_centers = (
        energies_edges[:-1] + (energies_edges[1:] - energies_edges[:-1]) / 2.0
    )
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[0.95, 0.05])
    cmap = plt.cm.get_cmap("viridis")
    cmap.set_under(plt.rcParams["axes.facecolor"])
    # Color map
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"Pairwise distance ($\mathrm{\AA}$)")
    ax.set_ylabel(r"Sum of $E_{PKA}$ (keV)")
    ax.set_ylim(energies_edges[[0, -1]])
    ax.set_xlim(dist_edges[[0, -1]])
    im = NonUniformImage(
        ax,
        cmap=cmap,
        extent=(*dist_edges[[0, -1]], *energies_edges[[0, -1]]),
    )
    im.set_clim(vmin=1 / nions)
    im.set_data(dist_centers, energies_centers, hist)
    ax.add_image(im)
    # Color bar
    cax = fig.add_subplot(gs[0, 1])
    fig.colorbar(im, cax, label="Counts per ion")
    # Finish
    fig.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()



"""This module contains the `XYZWriter` class."""

from dataclasses import dataclass, field
from io import TextIOWrapper
from pathlib import Path
from types import TracebackType
from typing import Optional

import numpy.typing as npt


@dataclass
class XYZWriter:
    """Class for writing structured data to an XYZ file format.

    Parameters
    ----------
    file_path : Path
        Path to the file where data will be written.
    mode : str, optional
        File open mode, by default "w".

    Attributes
    ----------
    file : TextIOWrapper
        File object associated with this writer.
    """

    file_path: Path
    mode: str = "w"
    file: Optional[TextIOWrapper] = field(default=None, init=False)

    def __post_init__(self) -> None:
        self.file = open(self.file_path, self.mode, encoding="utf-8")

    def __enter__(self) -> "XYZWriter":
        """Enter the runtime context related to this object."""
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        exc_traceback: Optional[TracebackType] = None,
    ) -> bool:
        """Exit the runtime context related to this object."""
        if self.file and not self.file.closed:
            self.file.close()
        return False

    def close(self) -> None:
        """Close the file associated with this writer."""
        if self.file and not self.file.closed:
            self.file.close()

    def __del__(self) -> None:
        file = getattr(self, "file", None)
        if file and not file.closed:
            file.close()

    def _get_properties(
        self, dtype: npt.DTypeLike
    ) -> tuple[tuple[str, ...], int, list[str], list[int]]:
        """Get the properties of the given data.

        Parameters
        ----------
        dtype : npt.DTypeLike
            Datatype of the given data.

        Returns
        -------
        tuple
            A tuple containing the names, count, types, and multiplicities of the properties.
        """
        name_props = dtype.names
        count_props = len(name_props)
        kind_map = {"i": "I", "f": "R", "U": "S"}
        type_props = []
        for descr in dtype.descr:
            kind = descr[1][1]
            if kind not in kind_map:
                raise TypeError(f"Unexpected dtype kind: {kind}")
            type_props.append(kind_map[kind])
        multiplicity_props = [
            dtype[name_prop].shape[0] if dtype[name_prop].shape else 1
            for name_prop in name_props
        ]

        return name_props, count_props, type_props, multiplicity_props

    def _get_comment(
        self,
        name_props: tuple,
        count_props: int,
        type_props: list,
        multiplicity_props: list,
    ) -> str:
        """Generate file comment following xyz guidelines.

        Parameters
        ----------
        name_props : tuple
            Property names.
        count_props : int
            Number of properties.
        type_props : list
            Property types.
        multiplicity_props : list
            Property multiplicities.

        Returns
        -------
        str
            Comment.
        """
        comment = "Properties=" + ":".join(
            f"{name_props[i]}:{type_props[i]}:{multiplicity_props[i]}"
            for i in range(count_props)
        )
        return comment

    def _data_to_line(
        self,
        data: npt.NDArray,
        name_props: tuple,
        count_props: int,
        multiplicity_props: list,
    ) -> str:
        """Transform data into string to write.

        Parameters
        ----------
        data : npt.NDArray
            Data to write.
        name_props : tuple
            Property names.
        count_props : int
            Number of properties.
        multiplicity_props : list
            Property multiplicities.

        Returns
        -------
        str
            Data as string.
        """
        return " ".join(
            (
                str(data[name_props[i]])
                if multiplicity_props[i] == 1
                else " ".join(map(str, data[name_props[i]]))
            )
            for i in range(count_props)
        )

    def save(self, datas: npt.NDArray, extra_comment: str = "") -> None:
        """
        Writes the given data into the file.

        Parameters
        ----------
        datas : npt.NDArray
            Data to write.
        extra_comment : str, optional
            Additional info to add at the end of the comment. Must follow xyz guidelines.
            Example: 'Info="Fe irradiatied in Fe, Ion 1"'
        """
        natoms = datas.size
        dtype = datas.dtype
        name_props, count_props, type_props, multiplicity_props = self._get_properties(
            dtype
        )
        comment = self._get_comment(
            name_props, count_props, type_props, multiplicity_props
        )
        full_comment = f"{comment} {extra_comment}" if extra_comment else comment
        self.file.write(f"{natoms}\n")
        self.file.write(f"{full_comment}\n")
        for data in datas:
            line = self._data_to_line(data, name_props, count_props, multiplicity_props)
            self.file.write(f"{line}\n")
